import hydra
import numpy as np
from plotly.offline import plot
import plotly.graph_objs as go
from omegaconf import OmegaConf
from os import path
from kamodo import Kamodo, compose
import os
from hydra import utils


def eval_config(params):
    """generate arrays dict from configuration"""
    args = {}
    if hasattr(params, 'items'):
        for k,v in params.items():
            try:
                args[k] = np.array(v)
            except:
                args[k] = np.linspace(v['min'], v['max'], v['n'])
    return args

def write_plot_div(plot_result, plot_filename):
    """writes plot div to file"""
    if plot_result is not None:
        with open(plot_filename, 'w') as f:
            f.write(plot_result)
            f.write('') # needs a newline or else embedding breaks

def config_override(cfg):
    """Overrides with user-supplied configuration

    kamodo will override its configuration using
    kamodo.yaml if it is in the current working directory
    or users can set an override config:
        config_override=path/to/myconfig.yaml
    """
    override_path = hydra.utils.to_absolute_path(cfg.config_override)
    if path.exists(override_path):
        override_conf = OmegaConf.load(override_path)
        # merge overrides first input with second
        cfg = OmegaConf.merge(cfg, override_conf)
    return cfg

def get_plot_dict(plot_params):
    if hasattr(plot_params, 'items'):
        pass
    elif type(plot_params) is str:
        plot_params = {plot_params: {}}
    else:
        try:
            plot_params = {k : {} for k in plot_params}
        except:
            raise IOError("cannot handle plot parameters of type {}".format(type(plot_params)))
    return plot_params


@hydra.main(config_path='conf/config.yaml', strict = False)
def main(cfg):
    """A low-coding command line interface for Kamodo

    This application allows users to work with kamodo-compatible models and data
    directly from the command line. 

    Custom models, data, and expressions may be composed by editing config files
    without needing to write python.
    """
    cfg = config_override(cfg)
    if cfg.verbose > 1:
        print(cfg.pretty())

    plot_conf = cfg.plot_conf
    plot_out_prefix = plot_conf.filename.split('.html')[0]
    plot_conf.pop('filename')

    models = dict()
    for model_name, model_conf in cfg.models.items():
        print(model_name)
        # model_conf['symbol_dict'] = locals()

        model = hydra.utils.instantiate(model_conf)
        models[model_name] = model

        if cfg.verbose > 0:
            print(model.detail())
        if cfg.verbose > 1:   
            print(model.to_latex())

        if model_conf.evaluate is not None:
            for varname, params in model_conf.evaluate.items():
                try:
                    units = model[varname].meta['units']
                    lhs = "{}({})".format(varname, ','.join(['{}={}'.format(k,v) for k,v in params.items()]))

                    rhs = model[varname](**eval_config(params))

                    print("{} {} = \n".format(lhs,units))
                    print(rhs)
                except Exception as m:
                    print('could not evaluate {} with params\n{}'.format(
                        varname, params))
                    print(m)
                    pass

        fig_layout = model_conf.fig_layout
        plot_params = model_conf.plot
        if plot_params is not None:
            plot_params = get_plot_dict(plot_params)

            for varname, params in plot_params.items():
                try:
                    plot_args = {varname: eval_config(params)}
                except:
                    print('issue with these params for {}'.format(varname))
                    print(params)
                    raise
                try:
                    fig = go.Figure(model.plot(**plot_args))
                    if fig_layout is not None:
                        fig.update_layout(**fig_layout)
                    out_filename = '{}_{}'.format(plot_out_prefix, varname)
                    plot_result = plot(fig, filename = out_filename, **plot_conf)
                    if plot_result is not None:
                        write_plot_div(plot_result, out_filename)
                except:
                    print('could not plot {} with params:'.format(varname))
                    print(plot_args)
                    print(plot_conf)
                    raise

    kamodo = compose(**models)
    if cfg.verbose > 0:
        print(kamodo.detail())
    if cfg.verbose > 1:
        print(kamodo.to_latex())


# entrypoint for package installer
def entry():
    main()

if __name__ == "__main__":
    main()


from kamodo import Kamodo
import pytest
import numpy as np
import pandas as pd
from .plotting import scatter_plot, line_plot, vector_plot, contour_plot, surface, plane, tri_surface_plot, get_arg_shapes, plot_types, plot_dict, image, symbolic_shape


def test_scatter_plot():
    title = 'test plot title'
    kamodo_obj = Kamodo(f='x**3')
    result = kamodo_obj.evaluate('f', x=np.random.random((100, 3)))
    titles = dict(variable='x', title=title)

    [trace], plot_type, layout = scatter_plot(result, titles, True)

    assert trace['mode'] == 'markers'
    assert layout['title']['text'] == title
    assert plot_type == '3d-scatter'


def test_line_plot_line():
    kamodo = Kamodo(f='x**3')
    result = kamodo.evaluate('f', x=np.linspace(-1, 1, 100))

    titles = dict(variable='f',
                  title='$f(x)=x^3$',
                  title_lhs='f', title_short='f')

    traces, plot_type, layout = line_plot(result, titles, True)

    assert traces[0]['name'] == 'f'
    assert layout['title']['text'] == '$f(x)=x^3$'
    assert plot_type == 'line'


def test_line_plot_2d_line():
    def fvec(t):
        x = t * np.sin(t) ** 3
        y = t * np.cos(t) * 2
        return np.vstack((x, y)).T

    title = '$\\vec{f}(t)=sin(t)^3, cos(t)^2 \\; t \\in [-\\pi, \\pi]$'
    kamodo = Kamodo(fvec=fvec)
    result = kamodo.evaluate('fvec',
                             t=np.linspace(-np.pi, np.pi, 100))

    titles = dict(variable='fvec',
                  title=title,
                  title_lhs='fvec',
                  title_short='f',
                  units='')

    traces, plot_type, layout = line_plot(result, titles, True)

    assert plot_type == '2d-line'
    assert layout['title']['text'] == title


def test_line_plot_3d_line_pd():
    def fvec(t):
        x = t * np.sin(t) ** 3
        y = t * np.cos(t) * 2
        z = t * np.sin(t)
        return pd.DataFrame(dict(X_GSE=x, Y_GSE=y, Z_GSE=z))

    title = '$\\vec{f}(t)=sin(t)^3, cos(t)^2 \\; t \\in [-\\pi, \\pi]$'
    kamodo = Kamodo(fvec=fvec)
    results = kamodo.evaluate('fvec',
                              t=np.linspace(-np.pi, np.pi, 100))

    titles = dict(variable='fvec',
                  title=title,
                  title_lhs='fvec',
                  title_short='f',
                  units='')

    traces, plot_type, layout = line_plot(results, titles, True)

    assert plot_type == '3d-line'
    assert layout['title']['text'] == title


def test_vector_plot_2d_vector():
    x = np.linspace(-np.pi, np.pi, 25)
    y = np.linspace(-np.pi, np.pi, 30)
    xx, yy = np.meshgrid(x, y)
    points = np.array(list(zip(xx.ravel(), yy.ravel())))

    def fvec_Ncomma2(rvec_Ncomma2=points):
        ux = np.sin(rvec_Ncomma2[:, 0])
        uy = np.cos(rvec_Ncomma2[:, 1])
        return np.vstack((ux, uy)).T

    title = '$\\vec{f}_{N,2}{\\left(\\vec{r}_{N,2} \\right)} = \\lambda{\\left(\\vec{r}_{N,2} \\right)}$'
    kamodo = Kamodo(fvec_Ncomma2=fvec_Ncomma2)
    results = kamodo.evaluate('fvec_Ncomma2', rvec_Ncomma2=points)

    titles = dict(variable='fvec_Ncomma2', title=title)
    [trace], plot_type, layout = vector_plot(results, titles, True, scale=0.1)

    assert trace['mode'] == 'lines'
    assert trace['name'] == 'quiver'
    assert plot_type == '2d-vector'
    assert layout['title']['text'] == title


def test_vector_plot_3d_vector():
    x, y, z = np.meshgrid(np.linspace(-2, 2, 4),
                          np.linspace(-3, 3, 6),
                          np.linspace(-5, 5, 10))
    points = np.array(list(zip(x.ravel(), y.ravel(), z.ravel())))

    def fvec_Ncomma3(rvec_Ncomma3=points):
        return rvec_Ncomma3

    title = '$\\vec{f}_{N,3}{\\left(\\vec{r}_{N,3} \\right)} = \\lambda{\\left(\\vec{r}_{N,3} \\right)}$'
    kamodo = Kamodo(fvec_Ncomma3=fvec_Ncomma3)

    results = kamodo.evaluate('fvec_Ncomma3', rvec_Ncomma3=points)

    titles = dict(variable='fvec_Ncomma3', title=title)
    [trace], plot_type, layout = vector_plot(results, titles, True)

    assert trace['hoverinfo'] == 'x+y+z+u+v+w+norm'
    assert plot_type == '3d-vector'
    assert layout['title']['text'] == title


def test_vector_plot_3d_vector_pd():
    x, y, z = np.meshgrid(np.linspace(-2, 2, 4),
                          np.linspace(-3, 3, 6),
                          np.linspace(-5, 5, 10))
    points = pd.DataFrame(list(zip(x.ravel(), y.ravel(), z.ravel())))

    def fvec_Ncomma3(rvec_Ncomma3=points):
        return rvec_Ncomma3

    title = '$\\vec{f}_{N,3}{\\left(\\vec{r}_{N,3} \\right)} = \\lambda{\\left(\\vec{r}_{N,3} \\right)}$'
    kamodo = Kamodo(fvec_Ncomma3=fvec_Ncomma3)

    results = kamodo.evaluate('fvec_Ncomma3', rvec_Ncomma3=points)

    titles = dict(variable='fvec_Ncomma3', title=title)

    [trace], plot_type, layout = vector_plot(results, titles, True)

    assert trace['hoverinfo'] == 'x+y+z+u+v+w+norm'
    assert plot_type == '3d-vector'
    assert layout['title']['text'] == title


def test_contour_plot_2d_grid():
    def f_NcommaM(x_N=np.linspace(0, 8 * np.pi, 100), y_M=np.linspace(0, 5, 90)):
        x, y = np.meshgrid(x_N, y_M, indexing='xy')
        return np.sin(x) * y

    title = '$\\operatorname{f_{N,M}}{\\left(x_{N},y_{M} \\right)} = \\lambda{\\left(x_{N},y_{M} \\right)}$'
    kamodo = Kamodo(f_NcommaM=f_NcommaM)

    results = kamodo.evaluate('f_NcommaM')
    titles = dict(variable='f_NcommaM', title=title)

    traces, chart_type, layout = contour_plot(results, titles, 'ij', True)

    assert chart_type == '2d-grid'
    assert layout['title']['text'] == title


def test_coutour_plot_ij_index_2d_grid():
    def f_NcommaM(x_N=np.linspace(0, 8 * np.pi, 100), y_M=np.linspace(0, 5, 100)):
        x, y = np.meshgrid(x_N, y_M, indexing='xy')
        return np.sin(x) * y

    title = '$\\operatorname{f_{N,M}}{\\left(x_{N},y_{M} \\right)} = \\lambda{\\left(x_{N},y_{M} \\right)}$'
    kamodo = Kamodo(f_NcommaM=f_NcommaM)

    results = kamodo.evaluate('f_NcommaM')
    titles = dict(variable='f_NcommaM', title=title)

    traces, chart_type, layout = contour_plot(results, titles, 'ij', True)

    assert chart_type == '2d-grid'
    assert layout['title']['text'] == title


def test_coutour_plot_xy_index_2d_grid():
    def f_NcommaM(x_N=np.linspace(0, 8 * np.pi, 100), y_M=np.linspace(0, 5, 100)):
        x, y = np.meshgrid(x_N, y_M, indexing='xy')
        return np.sin(x) * y

    title = '$\\operatorname{f_{N,M}}{\\left(x_{N},y_{M} \\right)} = \\lambda{\\left(x_{N},y_{M} \\right)}$'
    kamodo = Kamodo(f_NcommaM=f_NcommaM)

    results = kamodo.evaluate('f_NcommaM')
    titles = dict(variable='f_NcommaM', title=title)

    traces, chart_type, layout = contour_plot(results, titles, 'xy', True)

    assert chart_type == '2d-grid'
    assert layout['title']['text'] == title


def test_contour_plot_2d_skew():
    r = np.linspace(1, 3, 20)
    theta = np.linspace(0, np.pi, 14)
    r_, theta_ = np.meshgrid(r, theta)
    XX = r_ * np.cos(theta_)
    YY = r_ * np.sin(theta_)

    def f_NM(x_NM=XX, y_NM=YY):
        return np.sin(x_NM) + y_NM

    title = '$\\operatorname{f_{NM}}{\\left(x_{NM},y_{NM} \\right)} = \\lambda{\\left(x_{NM},y_{NM} \\right)}$'
    kamodo = Kamodo(f_NM=f_NM)

    results = kamodo.evaluate('f_NM')

    titles = dict(variable='f_NM', title=title)

    traces, chart_type, layout = contour_plot(results, titles, 'xy', True)

    assert chart_type == '2d-skew'
    assert layout['title']['text'] == title


def test_plane():
    def f_LMN(
            x_L=np.linspace(-5, 5, 50),
            y_M=np.linspace(0, 10, 75),
            z_N=np.linspace(-20, 20, 100)):
        xx, yy, zz = np.meshgrid(x_L, y_M, z_N, indexing='xy')
        return xx + yy + zz

    title = '$\\operatorname{f_{LMN}}{\\left(x_{L},y_{M},z_{N} \\right)} = \\lambda{\\left(x_{L},y_{M},z_{N} \\right)}$'
    kamodo = Kamodo(f_LMN=f_LMN)
    results = kamodo.evaluate('f_LMN', x_L=-5)
    titles = dict(variable='f_LMN', title=title)
    traces, chart_type, layout = plane(results, titles, verbose=True)

    assert chart_type == '3d-plane'
    assert layout['title']['text'] == title


def test_plane_meshgrid_false():
    def f_LMN(
            x_L=pd.DataFrame(np.linspace(-5, 5, 50)),
            y_M=pd.DataFrame(np.linspace(0, 10, 75)),
            z_N=pd.DataFrame(np.linspace(-20, 20, 75))):
        xx, yy, zz = np.meshgrid(x_L, y_M, z_N, indexing='xy')
        return xx + yy + zz

    title = '$\\operatorname{f_{LMN}}{\\left(x_{L},y_{M},z_{N} \\right)} = \\lambda{\\left(x_{L},y_{M},z_{N} \\right)}$'
    kamodo = Kamodo(f_LMN=f_LMN)
    results = kamodo.evaluate('f_LMN', x_L=-5)
    titles = dict(variable='f_LMN', title=title)

    traces, chart_type, layout = plane(results, titles, verbose=True)

    assert chart_type == '3d-plane'
    assert layout['title']['text'] == title


def test_surface_3d_surface():
    u = np.linspace(-2, 2, 40)
    v = np.linspace(-2, 2, 50)
    uu, vv = np.meshgrid(u, v)

    def parametric(x_NM=uu * np.sin(vv * np.pi),
                   y_NM=vv,
                   z_NM=np.exp(-uu ** 2 - vv ** 2)):
        return np.array([1])

    title = '$p{\\left(x_{NM},y_{NM},z_{NM} \\right)} = \\lambda{\\left(x_{NM},y_{NM},z_{NM} \\right)}$'
    kamodo = Kamodo(p=parametric)

    results = kamodo.evaluate('p')
    titles = dict(variable='p', title=title)

    traces, chart_type, layout = surface(results, titles)

    assert chart_type == '3d-surface'
    assert layout['title']['text'] == title


def test_surface_3d_surface_scalar():
    R = 1
    theta = np.linspace(.2 * np.pi, .8 * np.pi, 40)
    phi = np.linspace(0, 2 * np.pi, 50)
    theta_, phi_ = np.meshgrid(theta, phi)
    r = (R + .1 * (np.cos(10 * theta_) * np.sin(14 * phi_)))

    xx = r * np.sin(theta_) * np.cos(phi_)
    yy = r * np.sin(theta_) * np.sin(phi_)
    zz = r * np.cos(theta_)

    def spherelike(x_NM=xx, y_NM=yy, z_NM=zz):
        return .1 * x_NM + x_NM ** 2 + y_NM ** 2 + z_NM ** 2

    title = '$\\operatorname{h_{NM}}{\\left(x_{NM},y_{NM},z_{NM} \\right)} = \\lambda{\\left(x_{NM},y_{NM},z_{NM} \\right)}$'
    kamodo = Kamodo(h_NM=spherelike)

    results = kamodo.evaluate('h_NM')
    titles = dict(variable='h_NM', title=title)

    traces, chart_type, layout = surface(results, titles, True)

    assert chart_type == '3d-surface-scalar'
    assert layout['title']['text'] == title


def test_surface_len_of_result_variable_is_1():
    def f_LMN(
            x_L=np.linspace(-5, 5, 50),
            y_M=np.linspace(0, 10, 1),
            z_N=np.linspace(-20, 20, 1)):
        xx, yy, zz = np.meshgrid(x_L, y_M, z_N, indexing='xy')
        return xx + yy + zz

    title = '$\\operatorname{f_{LMN}}{\\left(x_{L},y_{M},z_{N} \\right)} = \\lambda{\\left(x_{L},y_{M},z_{N} \\right)}$'
    kamodo = Kamodo(f_LMN=f_LMN)
    results = kamodo.evaluate('f_LMN')
    titles = dict(variable='f_LMN', title=title)

    with pytest.raises(UnboundLocalError) as error:
        traces, chart_type, layout = surface(results, titles, True)

    assert "local variable 'chart_type' referenced before assignment" in str(error)


def test_tri_surface_plot():
    R = 1
    theta = np.linspace(.2 * np.pi, .8 * np.pi, 40)
    phi = np.linspace(0, 2 * np.pi, 50)
    theta_, phi_ = np.meshgrid(theta, phi)
    r = (R + .1 * (np.cos(10 * theta_) * np.sin(14 * phi_)))

    xx = r * np.sin(theta_) * np.cos(phi_)
    yy = r * np.sin(theta_) * np.sin(phi_)
    zz = r * np.cos(theta_)

    def tri_surface(x_NM=xx, y_NM=yy, z_NM=zz):
        return .1 * x_NM + x_NM ** 2 + y_NM ** 2 + z_NM ** 2

    title = '$\\operatorname{h_{NM}}{\\left(x_{NM},y_{NM},z_{NM} \\right)} = \\lambda{\\left(x_{NM},y_{NM},' \
            'z_{NM} \\right)}$ '
    kamodo = Kamodo(h_NM=tri_surface)

    results = kamodo.evaluate('h_NM')
    titles = dict(variable='h_NM', title=title)

    traces, chart_type, layout = tri_surface_plot(results, titles, True)

    assert chart_type == '3d-surface'
    assert layout['title']['text'] == title


def test_arg_shape_np():
    def fvec(t):
        x = t * np.sin(t) ** 3
        y = t * np.cos(t) * 2
        return np.vstack((x, y)).T

    kamodo = Kamodo(fvec=fvec)
    results = kamodo.evaluate('fvec',
                              t=np.linspace(-np.pi, np.pi, 100))

    f = results['fvec']

    shape = get_arg_shapes(f)

    assert type(f) == np.ndarray
    assert shape == [(100, 2)]


def test_arg_shape_pd():
    def fvec(t):
        x = t * np.sin(t) ** 3
        y = t * np.cos(t) * 2
        return pd.DataFrame(dict(X_GSE=x, Y_GSE=y))

    kamodo = Kamodo(fvec=fvec)
    results = kamodo.evaluate('fvec',
                              t=np.linspace(-np.pi, np.pi, 100))

    f = results['fvec']

    shape = get_arg_shapes(f)

    assert type(f) == pd.DataFrame
    assert shape == [(100, 2)]


def test_line_plot_len_of_result_3():
    r = np.linspace(1, 3, 20)
    theta = np.linspace(0, np.pi, 14)
    r_, theta_ = np.meshgrid(r, theta)
    XX = r_ * np.cos(theta_)
    YY = r_ * np.sin(theta_)

    def f_NM(x_NM=XX, y_NM=YY):
        return np.sin(x_NM) + y_NM

    title = '$\\operatorname{f_{NM}}{\\left(x_{NM},y_{NM} \\right)} = \\lambda{\\left(x_{NM},y_{NM} \\right)}$'
    kamodo = Kamodo(f_NM=f_NM)
    titles = dict(variable='f_NM', title=title)

    results = kamodo.evaluate('f_NM')

    [trace], chart_type, layout = line_plot(results, titles, True)

    assert trace['mode'] == 'lines'
    assert chart_type == '2d-line'
    assert layout['title'] == title


def test_line_plot_len_of_result_4():
    r = np.linspace(1, 3, 20)
    theta = np.linspace(0, np.pi, 14)
    r_, theta_ = np.meshgrid(r, theta)
    XX = r_ * np.cos(theta_)
    YY = r_ * np.sin(theta_)
    ZZ = r_ * np.sin(theta_)

    def f_NM(x_NM=XX, y_NM=YY, z_NM=ZZ):
        return np.sin(x_NM) + y_NM + z_NM

    title = '$\\operatorname{f_{NM}}{\\left(x_{NM},y_{NM},z_{NM} \\right)} = \\lambda{\\left(x_{NM},y_{NM},z_{NM} \\right)}$'
    kamodo = Kamodo(f_NM=f_NM)

    titles = dict(variable='f_NM', title=title)
    results = kamodo.evaluate('f_NM')

    [trace], chart_type, layout = line_plot(results, titles, True)

    assert trace['mode'] == 'lines'
    assert chart_type == '4d-line'
    assert layout['title']['text'] == title


def test_scatter_plot_pd():
    kamodo = Kamodo(f='x**3')
    r = pd.DataFrame({'x': np.linspace(0, 1, 100), 'y': np.linspace(-1, 1, 100), 'z': np.linspace(-1, 0, 100)})

    results = kamodo.evaluate('f', x=r)
    title = '$f{\\left(x \\right)} = x^{3}$'
    titles = dict(variable='f',
                  title=title,
                  title_lhs='f', title_short='f')

    [trace], plot_type, layout = scatter_plot(results, titles, True)

    assert trace['mode'] == 'markers'
    assert plot_type == '3d-scatter'
    assert layout['title']['text'] == title


def test_line_plot_3d_line():
    def fvec(t):
        x = t * np.sin(t) ** 3
        y = t * np.cos(t) * 2
        z = t * np.cos(t)
        return np.vstack((x, y, z)).T

    title = '$\\vec{f}{\\left(t \\right)} = \\lambda{\\left(t \\right)}$'
    kamodo = Kamodo(fvec=fvec)
    results = kamodo.evaluate('fvec',
                              t=np.linspace(-np.pi, np.pi, 100))

    titles = dict(variable='fvec',
                  title=title,
                  title_lhs='fvec',
                  title_short='f',
                  units='')

    [trace], chart_type, layout = line_plot(results, titles, True)
    assert trace['mode'] == 'lines'
    assert chart_type == '3d-line'
    assert layout['title']['text'] == title

def test_image_plot():
    def img(i=np.arange(33),j=np.arange(35)):
        ii, jj, kk = np.meshgrid(i,j,[100, 200, 255], indexing='ij')
        return kk

    kamodo = Kamodo(img=img)
    results = kamodo.evaluate('img')
    titles = dict(variable='img', title='mytitle')
    [trace], chart_type, layout = image(results, titles, True)


def test_plot_keys():
    for k in plot_types.to_dict(orient = 'index'):
        try:
            plot_dict[k[0]][k[1]]['name']
        except KeyError:
            print('could not find', k[0], k[1])
            raise


def test_symbolic_shape():
    assert symbolic_shape((3,1,3)) == ((3, 1, 3),)
    assert symbolic_shape((4,4,4)) == (('N', 'N', 'N'),)
    assert symbolic_shape((2,1,3)) == ((2, 1, 3),)
    assert symbolic_shape((4,5,6)) == (('N', 'M', 'L'),)
    assert symbolic_shape((4,5,1)) == (('N', 'M', 1),)
    assert symbolic_shape((5, 6), (6, 5), (5, 4)) == (('N', 'M'), ('M', 'N'), ('N', 'L'))




"""
Tests for kamodo.py

"""

import numpy as np
from sympy import symbols, Symbol
import pytest
from sympy.core.function import UndefinedFunction
import pandas as pd
from kamodo import Kamodo, get_unit, kamodofy, Eq
import functools
from sympy import lambdify, sympify
from kamodo import get_abbrev
from .util import get_arg_units
from .util import get_unit_quantity, convert_unit_to
from kamodo import from_kamodo, compose
from sympy import Function
from kamodo import KamodoAPI
from .util import serialize, NumpyArrayEncoder
from .util import get_kamodo_unit_system

import warnings


def test_Kamodo_expr():
    a, b, c, x, y, z = symbols('a b c x y z')
    kamodo = Kamodo(a=x ** 2, verbose=True)
    try:
        assert kamodo.a(3) == 3 ** 2
    except:
        print(kamodo.symbol_registry)
        raise


def test_Kamodo_ambiguous_expr():
    a, b, c, x, y, z = symbols('a b c x y z')
    with pytest.raises(NameError):
        kamodo = Kamodo()
        kamodo['a(b,c)'] = x ** 2 + y ** 2
    kamodo = Kamodo()
    kamodo['a(x, y)'] = x ** 2 + y ** 2
    assert kamodo.a(3, 4) == 3 ** 2 + 4 ** 2


def test_Kamodo_callable():
    kamodo = Kamodo(f=lambda x: x ** 2, verbose=False)
    assert kamodo.f(3) == 9
    kamodo = Kamodo(f=lambda x, y: x ** 2 + y ** 3)
    assert kamodo.f(3, 4) == 3 ** 2 + 4 ** 3


def test_Kamodo_callable_array():
    kamodo = Kamodo(f=lambda x: x ** 2)
    assert (kamodo.f(np.linspace(0, 1, 107)) == np.linspace(0, 1, 107) ** 2).all()


def test_Kamodo_str():
    kamodo = Kamodo('f(a,x,b) = a*x+b')
    try:
        assert kamodo.f(3, 4, 5) == 3 * 4 + 5
    except:
        print(kamodo.signatures)
        print(list(kamodo.keys()))
        raise


def test_Kamodo_latex():
    kamodo = Kamodo('$f(a,x,b) = a^x+b $')
    assert kamodo.f(3, 4, 5) == 3 ** 4 + 5


def test_Kamodo_get():
    kamodo = Kamodo('$f(a,x,b) = a^x+b $')
    try:
        assert kamodo['f'](3, 4, 5) == 3 ** 4 + 5
    except:
        print(kamodo.signatures)
        print(list(kamodo.keys()))
        raise


def test_Kamodo_set():
    kamodo = Kamodo()
    kamodo['f(a,x,b)'] = '$a^x+b$'
    assert kamodo.f(2, 3, 4) == 2 ** 3 + 4


def test_Kamodo_mismatched_symbols():
    with pytest.raises(NameError):
        kamodo = Kamodo('$f(a,b) = a + b + c$', verbose=False)
        assert 'f' not in kamodo


def test_Kamodo_assign_by_expression():
    kamodo = Kamodo()
    f, x, y = symbols('f x y')
    kamodo['f(x, y)'] = x ** 2 + y ** 2
    assert kamodo['f'](3, 4) == 3 ** 2 + 4 ** 2
    assert kamodo.f(3, 4) == 3 ** 2 + 4 ** 2


def test_Kamodo_assign_by_callable():
    kamodo = Kamodo(f=lambda x: x ** 2, verbose=False)
    assert kamodo['f'](3) == 3 ** 2
    assert kamodo.f(3) == 3 ** 2
    Kamodo(f=lambda x, y: x + y)

    def rho(x, y):
        return x + y

    kamodo['g'] = rho
    assert kamodo.g(3, 4) == 3 + 4


def test_Kamodo_composition():
    # f, g = symbols('f g', cls=UndefinedFunction)
    # x = Symbol('x')
    kamodo = Kamodo(f="$x^2$", verbose=True)
    kamodo['g(x)'] = "$f(x) + x^2$"
    kamodo['h(x)'] = 'g**2 + x**2'
    
    try:
        assert kamodo.f(3) == 3 ** 2
        assert kamodo.g(3) == 3 ** 2 + 3 ** 2
        assert kamodo.h(3) == (3 ** 2 + 3 ** 2) ** 2 + 3 ** 2
    except:
        print(kamodo)
        raise


def test_Kamodo_reassignment():
    a, b, c, x, y, z, r = symbols('a b c x y z r')
    kamodo = Kamodo('f(x) = 3*x+5', verbose=True)
    kamodo['r(x,y)'] = x + y
    kamodo['r(x,y)'] = "3*x + y"
    assert kamodo.r(1, 1) != 2
    assert kamodo.r(1, 1) == 4

def test_Kamodo_reassignment_units():
    kamodo = Kamodo(verbose=True)
    kamodo['s(x[km],y[km])[kg]'] = 'x + y'
    assert kamodo.s(1, 1) == 2
    kamodo['s(x[m],y[m])[g]'] = '3*x + y'
    assert kamodo.s(1, 1) == 4


def test_multivariate_composition():
    kamodo = Kamodo(f='x**2', g=lambda y: y ** 3, verbose=True)
    kamodo['h(x,y)'] = 'f(x) + g(y)'
    kamodo['i(x,y)'] = 'x + f(h(x,y))'
    assert kamodo.h(3, 4) == 3 ** 2 + 4 ** 3
    assert kamodo.i(3, 4) == 3 + (3 ** 2 + 4 ** 3) ** 2


def test_special_numbers():
    kamodo = Kamodo()
    kamodo['f'] = "${}^x$".format(np.e)
    assert np.isclose(kamodo.f(1), np.e)


def test_function_registry():
    kamodo = Kamodo("f(x) = x**2", "g(y) = y**3")
    kamodo['r(x,y)'] = "(x**2 + y**2)**(1/2)"
    kamodo['h(x,y)'] = 'f + g + r'
    assert 'h' in kamodo.signatures


def test_symbol_key():
    f = symbols('f', cls=UndefinedFunction)
    x = Symbol('x')
    f_ = f(x)
    kamodo = Kamodo(verbose=True)
    kamodo[f_] = x ** 2
    assert kamodo.f(3) == 9


def test_compact_variable_syntax():
    f = symbols('f', cls=UndefinedFunction)
    x = symbols('x')
    kamodo = Kamodo(f='x**2')  # f also registers f(x)
    kamodo['g(x)'] = 'f + x'
    assert kamodo.g(3) == 3 ** 2 + 3


def test_unit_registry():
    def rho(x, y):
        return x * y

    def v(x, y):
        return x + y

    v.meta = dict(units='km/s')
    kamodo = Kamodo(v=v, verbose=False)
    kamodo['rho[kg/cc]'] = rho
    try:
        assert kamodo.rho.meta['units'] == 'kg/cc'
        assert kamodo.v.meta['units'] == 'km/s'
    except:
        print('\n', pd.DataFrame(kamodo.signatures))
        raise

    with pytest.raises(NameError):
        kamodo['p[kg]'] = v
    assert 'p' not in kamodo


def test_to_latex():
    warnings.simplefilter('error')
    kamodo = Kamodo(f='x**2', verbose=True)
    assert str(kamodo.to_latex(mode='inline')) == r'$f{\left(x \right)} = x^{2}$'
    kamodo = Kamodo(g='x', verbose=True)
    assert str(kamodo.to_latex(mode='inline')) == r'$g{\left(x \right)} = x$'
    kamodo['f(x[cm])[kg]'] = 'x**2'
    kamodo['g'] = kamodofy(lambda x: x**2, units='kg', arg_units=dict(x='cm'), equation='$x^2$')
    kamodo['h'] = kamodofy(lambda x: x**2, units='kg', arg_units=dict(x='cm'))
    
    @kamodofy(units = 'kg/m^3', citation = 'Bob et. al, 2018')
    def rho(x = np.array([3,4,5]), y = np.array([1,2,3])):
        """A function that computes density"""
        return x+y
    kamodo['rho'] = rho
    kamodo.to_latex()


def test_expr_conversion():
    kamodo = Kamodo('$a[km] = x$', verbose=True)
    print(kamodo.items())
    kamodo.a

def test_get_unit_fail():
    with pytest.raises(NameError):
        get_unit('unregistered units')
    with pytest.raises(NameError):
        get_unit('runregistered')

def test_get_unit_quantity():
    mykm = get_unit_quantity('mykm', 'km', scale_factor=2)
    mygm = get_unit_quantity('mygm', 'gram', scale_factor=4)
    assert str(convert_unit_to(mykm, get_unit('m'))) == '2000*meter'
    assert str(convert_unit_to(mygm, get_unit('kg'))) == 'kilogram/250'

# def test_validate_units():
#     f, x = symbols('f x')
#     with pytest.raises(ValueError):
#         validate_units(Eq(f, x), dict(f=get_unit('kg'), x=get_unit('m')))
#     validate_units(Eq(f, x), dict(f=get_unit('kg'), x=get_unit('g')))

#     lhs_units = validate_units(sympify('f(x)'), dict(f=get_unit('kg'), x=get_unit('m')))
#     print(lhs_units)


def test_unit_conversion():
    kamodo = Kamodo('$a(x[m])[km/s] = x$',
                    '$b(y[cm])[m/s] = y$', verbose=True)
    kamodo['c(x[m],y[m])[m/s]'] = '$a + b$'
    assert kamodo.c(1, 2) == 1000 + 200


def test_get_unit():
    assert get_unit('kg/m^3') == get_unit('kg/m**3')


def test_unit_conversion_syntax():
    kamodo = Kamodo('rho[kg/m^3] = x', verbose=True)
    with pytest.raises(NameError):
        kamodo['d[kg]'] = 'rho'
        print(kamodo.detail())


def test_unit_composition():
    kamodo = Kamodo('m[kg] = x', verbose=True)
    kamodo['v[km/s]'] = 'y'
    kamodo['p(x,y)'] = 'm*v'
    try:
        assert get_unit(kamodo.signatures['p']['units']) == get_unit('kg*km/s')
    except:
        print(kamodo.signatures)
        raise

def test_unit_composition_mixed():
    kamodo = Kamodo('$rho[kg/m^3] = x^3$', '$v[cm/s] = y^2$', verbose=True)
    kamodo['p[Pa]'] = '$\\rho v^2$'


def test_unit_function_composition():
    kamodo = Kamodo('X[m] = x', verbose=True)

    @kamodofy(units='km/s', arg_units=dict(x = 'm'))
    def v(x):
        return x

    kamodo['v'] = v
    kamodo['speed'] = 'v(X)'
    assert kamodo.speed.meta['units'] == 'km/s'


def test_method_args():
    class TestModel(Kamodo):
        def __init__(self, *args, **kwargs):
            super(TestModel, self).__init__(*args, **kwargs)
            self['rho'] = self.density

        def density(self, alt, lat, lon):
            return alt + lat + lon

        density.meta = dict(units='1/cm^3')

    test = TestModel(verbose=True)
    assert str(list(test.keys())[0].args[0]) != 'self'

    test['r(alt, lat, lon)[1/m^3]'] = 'rho'
    try:  # check that args are in the correct order
        assert list(test.signatures.values())[0]['symbol'].args == list(test.signatures.values())[1]['symbol'].args
    except:
        print('\n', test.detail())
        raise


def test_komodofy_decorator():
    @kamodofy(units='kg/cm^3')
    def my_density(x, y, z):
        return x + y + z

    assert my_density.meta['units'] == 'kg/cm^3'
    assert my_density(3, 4, 5) == 12


def test_komodofy_register():
    @kamodofy(units='kg/cm^3')
    def my_density(x, y, z):
        return x + y + z

    kamodo = Kamodo(rho=my_density)
    assert kamodo.rho.meta['units'] == 'kg/cm^3'


def test_komodofy_method():
    class TestModel(Kamodo):
        def __init__(self, *args, **kwargs):
            super(TestModel, self).__init__(*args, **kwargs)
            self['rho'] = self.density

        @kamodofy(units='1/cm^3')
        def density(self, alt, lat, lon):
            return alt + lat + lon

    test = TestModel(verbose=False)
    assert test.density.meta['units'] == '1/cm^3'


#   try:
#       assert len(kamodo.detail()) == 1
#   except:
#       print kamodo.symbol_registry
#       print kamodo.detail()
#       raise

# # def test_remove_symbol():
# #     kamodo = Kamodo(f = 'x')
# #     kamodo['g'] = '2*f'
# #     kamodo.remove_symbol('f')
# #     try:
# #         assert len(kamodo.symbol_registry) == 1
# #         assert len(kamodo) == 2
# #     except:
# #         print '\n',kamodo.detail()
# #         raise

# def test_function_arg_ordering():
#   def x(r,theta,phi):
#       '''converts from spherical to cartesian'''
#       return r*np.sin(theta)*np.cos(phi)

#   kamodo = Kamodo(x = x)
#   rhs_args = get_function_args(kamodo.x)
#   lhs_args = kamodo.signatures.values()[0]['symbol'].args

#   for i in range(3):
#       assert lhs_args[i] == rhs_args[i]

# def test_function_change_of_variables():
#   def rho(x):
#       return x**2

#   kamodo = Kamodo(x = 'y')
#   kamodo['rho'] = rho
#   kamodo['rho_perp'] = 'rho(x)'

#   assert str(get_function_args(kamodo.rho_perp)[0]) == 'y'


def test_kamodofy_update_attribute():
    @kamodofy(units='cm', update='y')
    def f(x):
        return x # pragma: no cover

    kamodo = Kamodo(f=f)
    assert f.update == 'y'


def test_kamodo_coupling():
    @kamodofy(units='cm', update='y')
    def y_iplus1(y, x):
        return y + x

    @kamodofy(units='m')
    def x_iplus1(x, y):
        return x - y

    kamodo = Kamodo()
    kamodo['x_iplus1'] = x_iplus1
    kamodo['y_iplus1'] = y_iplus1

    kamodo.x_iplus1.update = 'x'
    assert kamodo.x_iplus1.update == 'x'
    assert kamodo.y_iplus1.update == 'y'

    simulation = kamodo.simulate(y=1, x=0, steps=1)

    for state in simulation:
        pass

    assert state['x'] == -1
    assert state['y'] == 0


def test_units():
    nT = get_unit('nT')
    assert str(nT) == 'nanotesla'
    assert str(nT.abbrev) == 'nT'

    R_E = get_unit('R_E')
    assert str(R_E) == 'earth radii'
    assert str(R_E.abbrev) == 'R_E'


def test_default_composition():
    ## Need to wrap function defaults
    def create_wrapped(args, expr):
        func = lambdify(args, expr)

        @functools.wraps(func)
        def wrapped(*_args, **kwargs):
            # Write the logic here to parse _args and kwargs for the arguments as you want them
            return func(*actual_args)

        return wrapped


def test_vectorize():
    @np.vectorize
    @kamodofy(units='kg')
    def f(x, y):
        return x + y

    kamodo = Kamodo(f=f)
    kamodo.f([3, 4, 5], 5)
    kamodo.evaluate('f', x=3,y=2)


def test_redefine_variable():
    kamodo = Kamodo(rho='x + y')
    kamodo['rho'] = 'a + b'
    kamodo['rho(a,b)'] = 'a*b'

def test_unit_composition_registration():
    server = Kamodo(**{'M': kamodofy(lambda r=3: r, units='kg'),
                       'V[m^3]': (lambda r: r**3)}, verbose=True)
    user = Kamodo(mass=server.M, vol=server.V,
              **{'rho(r)[g/cm^3]':'mass/vol'}, verbose=True)

    result = (3/3**3)*(1000)*(1/10**6)
    assert np.isclose(user.rho(3), result)


def test_unit_expression_registration():
    kamodo = Kamodo(verbose=True)
    kamodo['f(x[cm])[cm**2]'] = 'x**2'


def test_multi_unit_composition():
    kamodo = Kamodo('a(x[s])[km] = x', verbose=True)
    kamodo['b(x[cm])[g]'] = 'x'
    kamodo['c'] = 'b(a)'
    print(kamodo.c.meta)
    assert kamodo.c.meta['units'] == 'g'
    assert kamodo.c.meta['arg_units']['x'] == str(get_abbrev(get_unit('s')))

def test_unit_composition_conversion():
    kamodo = Kamodo('a(x[kg])[m] = x', verbose=True)
    kamodo['b(x[cm])[g]'] = 'x'
    kamodo['c'] = 'b(a)'

    assert kamodo.c.meta['units'] == 'g'
    assert kamodo.c.meta['arg_units']['x'] == 'kg'
    assert kamodo.c(3) == kamodo.b(100*kamodo.a(3))


def test_get_arg_units():

    def assign_unit(symbol, **units):
        if isinstance(symbol, str):
            symbol = sympify(symbol)
        return symbol.subs(units)

    f_units = assign_unit('f(x,y)', x=get_unit('s'), y=get_unit('hour'))
    g_units = assign_unit('g(a,b,c)', a=get_unit('km'), b=get_unit('cm'), c=get_unit('m'))
    unit_registry = {
        sympify('f(x,y)'): f_units,
        f_units: get_unit('kg/m^3'),
        sympify('g(a,b,c)'): g_units,
        g_units: get_unit('m^2')
    }
    x, c = symbols('x c')
    result = get_arg_units(sympify('h(g(f(x,y)))'), unit_registry)
    assert result[x] == get_unit('s')
    result = get_arg_units(sympify('f(x,y)*g(a,b,c)'), unit_registry)
    assert result[c] == get_unit('m')

def test_compose_unit_multiply():
    kamodo = Kamodo('a(x[kg])[m] = x', verbose=True)
    kamodo['e'] = '2*a'


def test_compose_unit_add():
    kamodo = Kamodo(verbose=True)

    @kamodofy(units='m', arg_units={'x': 'kg'})
    def a(x):
        return x

    kamodo['a'] = a
    kamodo['b(y[cm])[km]'] = 'y'
    kamodo['c(x,y)[km]'] = '2*a + 3*b'
    assert kamodo.c.meta['arg_units']['x'] == str(get_abbrev(get_unit('kg')))
    assert kamodo.c.meta['arg_units']['y'] == str(get_abbrev(get_unit('cm')))
    result = 2*(3)/1000 + 3*(3)
    assert kamodo.c(3, 3) == result

def test_compose_unit_raises():

    with pytest.raises(NameError):
        kamodo = Kamodo('a(x[kg])[m] = x',
                        'b(y[cm])[km] = y', verbose=True)

        # this should fail since x is registered with kg
        kamodo['c(x[cm],y[cm])[km]'] = '2*a + 3*b'


def test_repr_latex():
    kamodo = Kamodo(f='x')
    assert kamodo._repr_latex_() == r'\begin{equation}f{\left(x \right)} = x\end{equation}'
    kamodo = Kamodo(f=lambda x:x)
    assert kamodo.f._repr_latex_() == '$f{\\left(x \\right)} = \\lambda{\\left(x \\right)}$'


def test_dataframe_detail():
    kamodo = Kamodo(f='x')
    type(kamodo.detail())
    assert len(kamodo.detail()) == 1
    assert isinstance(kamodo.detail(), pd.DataFrame)

def test_from_kamodo():
    kamodo = Kamodo(f='x')
    knew = from_kamodo(kamodo, g='f**2')
    assert knew.g(3) == 9


def test_jit_evaluate():
    """Just-in-time evaluation"""
    kamodo = Kamodo(f='x')
    assert kamodo.evaluate('g=f**2', x = 3)['x'] == 3
    with pytest.raises(SyntaxError):
        kamodo.evaluate('f**2', x = 3)['x'] == 3

def test_eval_no_defaults():
    kamodo = Kamodo(f='x', verbose=True)
    kamodo['g'] = lambda x=3: x
    kamodo['h'] = kamodofy(lambda x=[2,3,4]: (kamodofy(lambda x_=np.array([1]): x_**2) for x_ in x))
    assert kamodo.evaluate('f', x=3)['x'] == 3
    assert kamodo.evaluate('g')['x'] == 3
    assert kamodo.evaluate('h')['x_'][-2] == 1.

def test_compose():
    k1 = Kamodo(f='x')
    k2 = Kamodo(g='y**2', h = kamodofy(lambda x: x**3))
    k3 = compose(m1=k1, m2=k2)
    assert k3.f_m1(3) == 3
    assert k3.g_m2(3) == 3**2
    assert k3.h_m2(3) == 3**3
    k3['h(f_m1)'] = 'f_m1'
    assert k3.h(3) == 3

def test_symbol_replace():
    k = Kamodo(f='x', verbose=True)

    f1, f2 = list(k.keys())
    print('\n|||||',f1, f2, '||||||')
    k[f1] = 'x**2'
    assert k.f(3) == 9
    print('\n|||||', *list(k.keys()), '|||||')
    k[f2] = 'x**3'
    print('\n|||||', *list(k.keys()), '|||||')
    assert k.f(3) == 27

def test_contains():
    kamodo = Kamodo(f='x', verbose=True)
    assert 'f' in kamodo
    assert 'f( x )' in kamodo
    f, x = symbols('f x')
    assert f in kamodo # f is a symbo
    f = Function(f) # f is an Unefined Function
    assert f in kamodo
    assert f(x) in kamodo
    assert f('x') in kamodo

def test_unusual_signature():
    with pytest.raises(NotImplementedError):
        kamodo = Kamodo()
        kamodo['f(x)=f(cm)=kg'] = 'x'
    with pytest.raises(NotImplementedError):
        kamodo = Kamodo('f(x)=f(cm)=kg=x')


# def test_remove_symbol():
#     kamodo = Kamodo(f='x', verbose=True)
#     kamodo.remove_symbol('f')
#     assert 'f' not in kamodo

def test_method_registry():
    
    class MyClass(Kamodo):
        @kamodofy
        def f(self, x):
            return x**2

    myclass = MyClass()
    myclass['f'] = myclass.f

def test_del_function():
    kamodo = Kamodo(f='x', g='y', h='y', verbose=True)
    del(kamodo.f)
    assert 'f' not in kamodo
    assert 'f' not in kamodo.signatures
    del(kamodo['g'])
    assert 'g' not in kamodo
    del(kamodo['h(y)'])
    print(kamodo.keys())
    assert 'h(y)' not in kamodo

    with pytest.raises(AttributeError):
        del(kamodo.y)

def test_simple_figure():
    @kamodofy(units='kg', hidden_args=['ions'])
    def f_N(x_N):
        return x_N**2

    kamodo = Kamodo(f_N=f_N,verbose=True)
    kamodo.plot(f_N=dict(x_N=np.linspace(-4, 3, 30)))

def test_unavailable_4d_plot_type():
    def g(x=np.array([1]),
          y=np.array([1]),
          z=np.array([1]),
          t=np.array([1])):
        return x**2 + y**2 + z**2 + t**2

    kamodo = Kamodo(g=g, verbose=True)
    with pytest.raises(KeyError):
        kamodo.plot('g')

def test_multiple_traces():
    kamodo = Kamodo(f='x', g='x**2')
    kamodo.plot(
        f=dict(x=np.linspace(-1, 1, 10)),
        g=dict(x=np.linspace(-5, 5, 10)))

def test_unitless_composition():
    @kamodofy
    def alpha(x):
        return x

    @kamodofy
    def beta(y):
        return y


    kamodo = Kamodo(alpha=alpha, beta_=beta, verbose=True)
    kamodo['Gamma'] = 'alpha(beta_)'
    kamodo

def test_reserved_name():
    kamodo = Kamodo(verbose=True)
  
    @kamodofy
    def test(x, y):
        return x+y
    with pytest.raises(NotImplementedError):
        kamodo['test'] = test


class Ktest(Kamodo):
    def __init__(self, **kwargs):
        super(Ktest, self).__init__()

        t_N = pd.date_range('Nov 9, 2018', 'Nov 20, 2018', freq = 'H')
        @kamodofy(units='kg/m^3')
        def rho_N(t_N=t_N):
            t_N = pd.DatetimeIndex(t_N)
            t_0 = pd.to_datetime('Nov 9, 2018') 
            try:
                dt_days = (t_N - t_0).total_seconds()/(24*3600)
            except TypeError as err_msg:
                return 'cannot work with {} {}  {}'.format(type(t_N), type(t_N[0]), err_msg)

            result = np.abs(weierstrass(dt_days))
            return result

        @kamodofy(units='nPa')
        def p(x=np.linspace(-5, 5, 30)):
            try:
                return x**2
            except TypeError as m:
                print(m)
                print(type(x), x[0])
                raise

        @kamodofy(
            equation=r"\sum_{n=0}^{500} (1/2)^n cos(3^n \pi x)",
            citation='https://en.wikipedia.org/wiki/Weierstrass_function'
            )
        def weierstrass(x = np.linspace(-2, 2, 1000)):
            '''
            Weierstrass  function
            A continuous non-differentiable 
            https://en.wikipedia.org/wiki/Weierstrass_function
            '''
            nmax = 500
            n = np.arange(nmax)

            xx, nn = np.meshgrid(x, n)
            ww = (.5)**nn * np.cos(3**nn*np.pi*xx)
            return ww.sum(axis=0)
                

        self['rho_N'] = rho_N
        self['p'] = p
        self['Weierstrass'] = weierstrass


def test_kamodo_inline_merge():
    k1 = Kamodo(f='x**2')
    k2 = Kamodo(g=lambda y: y-1)

    # create a local namespace holding both kamodo objects
    ns = {'k1':k1, 'k2': k2}
    k3 = Kamodo(myf = sympify('k1.f(x) + k2.g(y)', locals=ns))
    assert k3.myf(x=3, y=4) == 3**2 + 4 - 1

def test_default_forwarding():
    x = np.linspace(-5, 5, 12)
    
    def f(x=x):
        return x**2
    
    k = Kamodo(f=f)
    k['g'] = 'f+2'
    assert len(k.g()) == 12
    
def test_multi_arg_units():
    kamodo = Kamodo(verbose=True)

    # @kamodofy(units='m', arg_units={'X': 'kg', 'Y':'cm', 'Z': 's'})
    # def f(X, Y, Z):
    #     return x*y*z

    kamodo['f(X[kg],Y[cm],Z[s])[m]'] = 'X*Y*Z'
    kamodo['a[g]'] = 'x'
    kamodo['b[m]'] = 'y'
    kamodo['c[ms]'] = 'z**2'
    kamodo['d(x,y,z)[cm]'] = 'f(a,b,c)'

def test_broken_unit():
    k = Kamodo()
    k['f[N]'] = 'x'

    get_unit('newton')
    get_unit('N')

def test_frequency_composition():
    @kamodofy(units='rad/s', arg_units={'B':'T', 'n_e':'1/m**3'})
    def omega_uh1(B, n_e):
        return np.sqrt(B**2+n_e**2)


    kamodo_test = Kamodo(verbose=True)
    kamodo_test['B_mag'] = kamodofy(lambda B=np.linspace(0.1,1.,10): B, units='nT', arg_units={'B':'nT'})
    kamodo_test['n_e'] = kamodofy(lambda n=np.linspace(4.,13.,10)*10**19:n, units='1/m**3', arg_units={'n':'1/m**3'})
    kamodo_test['omega_uh1'] = omega_uh1
    print(kamodo_test.unit_registry)

    #---------(input)--------
    kamodo_test['omega_uh1A'] = 'omega_uh1(B_mag, n_e)'
    kamodo_test.omega_uh1A


def test_frequency_units():
    omega = get_unit('rad')/get_unit('s')
    freq = get_unit('deg')/get_unit('s')
    kamodo_units = get_kamodo_unit_system()
    convert_unit_to(omega, freq, kamodo_units)




import hydra
from hydra.experimental import compose
from omegaconf import OmegaConf
import dash
import dash_core_components as dcc
import dash_html_components as html
from kamodo.cli.main import eval_config
from kamodo import get_defaults
import plotly.graph_objs as go
from dash.dependencies import Input, Output, ClientsideFunction, State
import numpy as np
from os import path
import os
import numpy as np
from sympy.core.function import UndefinedFunction 
import pandas as pd
import sys 
from dash_katex import DashKatex
from plotly.subplots import make_subplots
from dash.exceptions import PreventUpdate
from flask_restful import reqparse, abort, Api, Resource
import flask
import json
from kamodo import getfullargspec
from flask.json import jsonify
from flask import request
from io import StringIO


try:
    hydra.experimental.initialize(strict=False)
except:
    pass

from ast import literal_eval


def get_equation_divs(model, model_name, model_params):

    equations = []

    options = []
    selected = []
    for var_symbol, func in model.items():
        if type(var_symbol) != UndefinedFunction:
            var_label = str(type(var_symbol))

            # make sure user-specified params get a check mark
            if str(type(var_symbol)) in model_params:
                selected.append(var_label)

            var_docs = model[var_symbol].__doc__.split('\n')
            doc_summary = var_docs[0]
            doc_complete = ''
            if len(var_docs) > 1:
                doc_complete = [html.Pre(doc_) for doc_ in var_docs[1:]]
            options.append({'label': '', 'value': var_label})
            equations.append(
                html.Div([
                    html.Div(
                        DashKatex(
                            id = "{}-{}-expression".format(model_name, str(var_symbol)),
                            expression = "{}".format(model.to_latex(
                                keys = [str(var_symbol)],
                                mode = "plain")),
                            ),
                        className = 'four columns'),
                    
                    html.Details([
                        html.Summary(doc_summary),
                        html.Div(children=doc_complete)],
                        className = 'eight columns',
                        style = {'text-align': 'left'})
                    ],
                    className = 'row'
                    )
                )

    equation_divs = html.Div([
            dcc.Checklist(
                id = 'checklist-{}'.format(model_name),
                options = options,
                value = selected,
                className = "one columns",
                style = {'text-align': 'right'},
                ),
            html.Div(
                children = equations,
                className = "eleven columns"
                ),
            ],
        className = 'row',
        )

    return equation_divs

def get_gui(cfg, server):

    app = dash.Dash(__name__,
        server=server,
        routes_pathname_prefix = cfg.gui.routes_pathname_prefix,
        external_stylesheets = cfg.gui.external_stylesheets,
        external_scripts = cfg.gui.external_scripts,
        )
    app.title = "kamodo"
    app.config.suppress_callback_exceptions = False

    tabs_children, graphs = get_tabs_children(cfg, app)

    app.layout = html.Div(children=[
        html.Div(
            [
            html.Div(
                html.Img(
                src="assets/Kamodo-black.png",
                alt="Kamodo Software Suite",
                style={'height':'80px'},
                ),
                className = 'six columns'
                ),
            html.Div(
                html.Img(
                src="assets/ccmc_logo.svg",
                alt="Kamodo Software Suite",
                style={'height':'80px'},
                ),
                className = 'six columns',
                style = {'text-align': 'right'}, 
                ),
            ],
            className = 'row'
            ),
        

        html.Div(children='''
            A low-coding interface for models and data
        '''),
        html.Div(children = [
            dcc.Textarea(
                id = "kamodo-config",
                title = "Configuration for Kamodo",
                placeholder = 'Supply a configuration for Kamodo',
                value = "{}".format(cfg.pretty()),
                style = {'width': '500px', 'height':'400px'},
                className = 'three columns'),
            html.Div(
                id = 'config-render',
                className = 'three columns',
                style = {'height': '400px'}
                ),
            dcc.Input(
                id = 'save-as',
                value = cfg.config_override,
                placeholder = "enter filename",
                className = 'two columns'
                ),
            html.Button(
                'Save',
                id = 'save-button',
                n_clicks = 0,
                )
            ],
            className = 'row',
            style = {'overflow': 'auto'}
        ),
        dcc.Tabs(id = 'kamodo-tabs',
            children = tabs_children),
    ])


    generate_config_render_callback(app)

    generate_kamodo_tabs_callback(app)

    generate_save_button_callbacks(app)

    for graph_id, (model, model_params, model_name) in graphs.items():
        generate_subplot_callback(app, graph_id, model, model_name, model_params)


    
    return app

def generate_save_button_callbacks(app):
    @app.callback(
        Output('save-button', 'children'),
        [Input('save-button', 'n_clicks'), ],
        [State('save-as', 'value'), State('kamodo-config', 'value')]
        )
    def save_config(n_clicks, filename, conf):
        if n_clicks > 0:
            try:
                cfg = OmegaConf.create(conf)
                with open(filename, 'w') as f:
                    f.write(cfg.pretty())
                    return 'saved'
            except:
                return html.Div('error', style = {'color':'red'})
        else:
            raise PreventUpdate
            return 'save'

    @app.callback(
        Output('save-as','value'),
        [Input('kamodo-config', 'value')])
    def update_save_value(conf):
        try:
            cfg = OmegaConf.create(conf)
            filename = cfg.config_override
            return filename
        except:
            raise PreventUpdate

    @app.callback(
        Output('save-button', 'n_clicks'),
        [Input('kamodo-config', 'value')])
    def update_save_button(conf):
        try:
            OmegaConf.create(conf)
            return 0
        except:
            raise PreventUpdate



def generate_config_render_callback(app):
    @app.callback(
        Output('config-render', 'children'),
        [Input('kamodo-config', 'value')])
    def config_render(conf):
        try:
            cfg = OmegaConf.create(conf)
            return dcc.Markdown(
                id = 'config-render-markdown',
                children = "```yaml\n{}\n```".format(cfg.pretty()),
                )
        except Exception as m:
            return html.Div(str(m), style = {'color':'red'})



def get_tabs_children(cfg, app):

    tabs_children = []
    graphs = {}

    for model_name, model_conf in cfg.models.items():
        print(model_name)
        tab_children = []
        tab_children.append(html.Br())

        try:
            model = hydra.utils.instantiate(model_conf)
        except Exception as m:
            print('could not instantiate {}'.format(model_name))
            raise

        if len(model) == 0:
            continue

        model_params = {}
        if model_conf.plot is None:
            for var_symbol, func in model.items():
                if type(var_symbol) == UndefinedFunction:
                    model_params[str(var_symbol)] = get_defaults(func)
        else:
            for var_name, var_args in model_conf.plot.items():
                if var_args is None:
                    try:
                        model_params[var_name] = get_defaults(model[var_name])
                    except Exception as m:
                        print(model.detail)
                        print(m)
                        sys.exit()
                else:
                    model_params[var_name] = eval_config(var_args)
     

        tab_children.append(get_equation_divs(model, model_name, model_params))

        graph_id = "graph-{}".format(model_name)

        tab_children.append(
            html.Div(
                html.Div(
                    dcc.Graph(
                        id = graph_id,
                        # figure = fig_subplots,
                        style = {
                            'height': '1000px', 
                            'width':'1200px', 
                            'text-align': 'center',
                            'margin': 'auto'}),
                    className = 'twelve columns'),
                className = 'row'),
            )

        try:
            plots = get_kamodo_plots(model, model_params.keys(), model_params)
        except Exception as m:

            tab_children.append(html.Div(str(m), style = {'color': 'red'}))
            plots = {}
                    
        tab_children.append(
            dcc.Store(id = graph_id + '-store',
                data = plots),
            )



        # fig_subplots = make_kamodo_subplots(model, model_params.keys(), model_params)
 



        graphs[graph_id] = model, model_params, model_name

        tab = dcc.Tab(label = model_name, children = tab_children)
        
        tabs_children.append(tab)

    return tabs_children, graphs

def generate_kamodo_tabs_callback(app):
    @app.callback(Output('kamodo-tabs','children'),
    [Input('kamodo-config', 'value')])
    def kamodo_content(conf):
        try:
            cfg = OmegaConf.create(conf)
            tabs_children, graphs = get_tabs_children(cfg, app)
            return tabs_children
        except Exception as m:
            print("could not get tabs children")
            print(m)
            print(conf)
            return html.Pre(str(m).split('\n'), style = {'color':'red'})
            

def get_kamodo_plots(model, var_names, model_params):
    plots = {}
    for var_symbol, func in model.items():
        if type(var_symbol) != UndefinedFunction:
            var_label = str(type(var_symbol))

            fig = model.figure(
                var_label,
                indexing = 'ij',
                return_type = True, 
                **model_params.get(var_label, {}))
            plots[var_label] = fig['data'][0]

    return plots


# def make_kamodo_subplots(model, var_names, model_params):
#     fig_subplots = make_subplots(rows=len(var_names), cols=1)
#     for plot_count, varname in enumerate(var_names):
#         try:
#             fig = model.figure(
#                 varname,
#                 indexing = 'ij',
#                 return_type = True, 
#                 **model_params.get(varname, {}))
#             fig_subplots.add_trace(fig['data'][0], row = plot_count+1, col = 1)
#         except:
#             print("could not plot {} with model_params:".format(varname))
#             print(model_params)
#             raise PreventUpdate
#     return fig_subplots


def generate_subplot_callback(app, graph_id, model, model_name, model_params):
    app.clientside_callback(
        ClientsideFunction("clientside", "subplots"),
        Output(component_id=graph_id, component_property="figure"),
        [Input(graph_id + '-store', "data"),
        Input('checklist-{}'.format(model_name), 'value')],
    )


    

def generate_callbacks(app, graphs):
    print("generating callbacks for {} graphs".format(len(graphs)))
    for graph_id, (chart_type, varname, plot_args, model, graph) in graphs.items():
        if chart_type == 'line':
            generate_range_slider_callback(
                app,
                graph_id,
                varname,
                model,
                **plot_args)
        else:
            print("cannot generate callback for chart_type {}".format(chart_type))

def generate_range_slider_callback(app, graph_id, varname, model, **plot_kwargs):
    print("generating calback for {}: {}".format(graph_id, varname))
    output = Output(graph_id, "figure")
    input_dict = dict()
    for k,v in plot_kwargs.items():
        input_dict[k] = [
            Input("rangeslider-{}-{}".format(graph_id, k), "value"),
            Input("input-{}-{}".format(graph_id, k), "value")]

    inputs  = []
    for inputs_ in input_dict.values():
        inputs.extend(inputs_)

    @app.callback(output, inputs)
    def update_figure(*inputs):
        plot_args = dict()
        for i,k in enumerate(input_dict):
            plot_args[k] = np.linspace(*inputs[2*i], num = inputs[2*i+1])

        fig = model.figure(
                varname,
                indexing = 'ij',
                return_type = True, 
                **plot_args)
        return go.Figure(fig)


def get_range_slider(graph_id, **kwargs):
    divs = []
    for k,v in kwargs.items():
        try:
            print('type is: {}'.format(type(v[0])))
            marks = {i: "{0:0.2f}".format(i) for i in v[::int(len(v)/31)]}
        except:
            marks = {i: str(i) for i in v[::int(len(v)/31)]}

        # print(len(marks))
        # print(marks)

        row_children = [
            html.Div(
                className = "eight columns",
                children = [
                    dcc.RangeSlider(
                        id = "rangeslider-{}-{}".format(graph_id, k),
                        marks=marks,
                        min=min(v),
                        max=max(v),
                        step = (max(v) - min(v))/len(v),
                        value=[min(v), max(v)],
                        updatemode='drag',
                        )]),
            html.Div(
                className = "four columns",
                children = [
                    dcc.Input(
                        id = "input-{}-{}".format(graph_id, k),
                        placeholder='number of points in {}'.format(k),
                        type='number',
                        value = len(v),
                        min = 3,
                        max = 10*len(v))])
            ]

        row_div = html.Div(
            className = 'row',
            children = row_children)
        divs.append(row_div)

    return divs


def config_override(cfg):
    """Overrides with user-supplied configuration

    hourly will override its configuration using
    hourly.yaml if it is in the current working directory
    or users can set an override config:
        config_override=path/to/myconfig.yaml
    """
    if cfg.config_override is not None:
        override_path = hydra.utils.to_absolute_path(cfg.config_override)
        if path.exists(override_path):
            override_conf = OmegaConf.load(override_path)
            # merge overrides first input with second
            cfg = OmegaConf.merge(cfg, override_conf)
    return cfg

def get_api(app):
    api = Api(app)
    return api


def main():

    cfg = compose('conf/kamodo.yaml')

    cli_conf = OmegaConf.from_cli()
    
    cfg = OmegaConf.merge(cfg, cli_conf)
        
    config_override = None
    extra_files = []
    if cfg.config_override is not None:
        override_path = "{}/{}".format(os.getcwd(),cfg.config_override)
        if path.exists(override_path):
            print("found {}".format(override_path))
            config_override = OmegaConf.load(override_path)
            # print(config_override.pretty())
            extra_files.append(override_path)
        else:
            if cfg.verbose > 0:
                print("could not get override: {}".format(override_path))
    

        if config_override is not None:
            cfg = OmegaConf.merge(cfg, config_override)
    

    if cfg.verbose > 0:
        print(cfg.pretty())

    server = flask.Flask(__name__)

    # app = Flask(__name__)
    api = Api(server)

    models = dict()
    for model_name, model_conf in cfg.models.items():
        try:
            # this might fail
            model_ = hydra.utils.instantiate(model_conf)
            models[model_name] = model_
        except:
            pass
    
    class Models(Resource):
        def get(self):
            details = dict()
            for model_name, model_ in models.items():
                detail = model_.detail().astype(str)
                print(detail)
                details[model_name] = detail.to_dict(
                    # default_handler=str, 
                    # indent =4, 
                    orient = 'index',
                    )
            return details

    api.add_resource(Models, '/api', '/api/')


    for model_name, model_ in models.items():
        api.add_resource(
            get_model_resource(model_name, model_),
            '/api/{}'.format(model_name),
            '/api/{}/'.format(model_name),
            endpoint = model_name)

        for var_symbol in model_:
            if type(var_symbol) != UndefinedFunction:
                var_label = str(type(var_symbol))
                api.add_resource(
                    get_func_resource(model_name, model_, var_symbol),
                    '/api/{}/{}'.format(model_name, var_label),
                    endpoint =  '/'.join([model_name, var_label]))



    @server.route('/')
    def index():
        return 'Hello Flask app'

    app = get_gui(cfg, server)
    app.run_server(debug = True, extra_files = extra_files)


def get_model_resource(model_name, model):
    class modelResource(Resource):
        def get(self):
            return model.detail().astype(str).to_dict(orient='index')

    return modelResource

def get_func_resource(model_name, model, var_symbol):
    parser = reqparse.RequestParser()
    func = model[var_symbol]
    defaults = get_defaults(func)
    for arg in getfullargspec(func).args:
        if arg in defaults:
            parser.add_argument(arg, type = str)
        else:
            parser.add_argument(arg, type = str, required = True)

    class funcResource(Resource):
        def get(self):
            
            args_ = parser.parse_args(strict = True)
            args = dict()

            for argname, val_ in args_.items():
                args[argname] = pd.read_json(StringIO(val_), typ = 'series')

            return func(**args).tolist()
      
            

    return funcResource

# entrypoint for package installer
def entry():
    main()


if __name__ == "__main__":
    main()



from collections import OrderedDict

import numpy as np
import pytest
from sympy import sympify as parse_expr, Symbol, symbols
from sympy.core.function import UndefinedFunction

from kamodo import Kamodo
from kamodo.util import kamodofy, gridify, sort_symbols, valid_args, eval_func, get_defaults, cast_0_dim, \
    concat_solution, get_unit_quantity, substr_replace, beautify_latex, arg_to_latex, simulate, pad_nan, \
    pointlike, solve, event, is_function

from .util import serialize, deserialize
import pandas as pd
import json
from .util import LambdaGenerator
from .util import curry, partial, get_args

@kamodofy
def rho(x=np.array([3, 4, 5])):
    """rho means density"""
    return x ** 2


@kamodofy
def divide(x=np.array([3, 4, 5])):
    """rho means density"""
    return x / 2


@np.vectorize
def myfunc(x, y, z):
    return x + y + z


def defaults_fun(var1, var2, default_var1=10, default_var2=20):
    return var1 + var2 + default_var1 + default_var2


@kamodofy
@gridify(x=np.linspace(-3, 3, 20),
         y=np.linspace(-5, 5, 10),
         z=np.linspace(-7, 7, 30))
def grid_fun(xvec):
    """my density function, returning array of size N

    xvec should have shape (N,3)"""
    return xvec[:, 0]


@pointlike(squeeze='0')
def squeeze_error_point(x=np.linspace(0, 8 * np.pi, 100)):
    return np.sin(x)


@pointlike(squeeze=0)
def squeeze_point(x=np.linspace(0, 8 * np.pi, 100)):
    return np.sin(x)


@pointlike()
def points(x=np.linspace(0, 8 * np.pi, 100)):
    return np.sin(x)


def fprime(x):
    return x * x


@event
def boundry(x, s=np.array([1, 2])):
    r = np.linalg.norm(s)

    if np.isnan(r):
        result = 0
    else:
        result = r - 1
    return result


def test_kamodofy():
    comparison = rho.data == np.array([9, 16, 25])
    assert comparison.all()
    comparison = divide(np.array([6, 13, 2])) == np.array([3, 13 / 2, 1])
    assert comparison.all()


def test_sort_symbols():
    myexpr = parse_expr('x+rho+a+b+c')
    assert sort_symbols(myexpr.free_symbols) == (Symbol('a'), Symbol('b'), Symbol('c'), Symbol('rho'), Symbol('x'))


def test_valid_args():
    assert valid_args(myfunc, dict(x='a', other='you')) == OrderedDict([('x', 'a')])
    assert valid_args(myfunc, dict(y='b', other='you')) == OrderedDict([('y', 'b')])
    assert valid_args(myfunc, dict(z='c', other='you')) == OrderedDict([('z', 'c')])
    assert valid_args(myfunc, dict(x='a', y='b', z='c', other='you')) == OrderedDict(
        [('x', 'a'), ('y', 'b'), ('z', 'c')])


def test_eval_func():
    def f(x):
        return x + 'a'

    with pytest.raises(TypeError) as error:
        eval_func(f, dict(x=1))

    assert eval_func(myfunc, dict(x='a', y='b', z='c', other='you')) == myfunc('a', 'b', 'c')
    assert eval_func(myfunc, dict(x=1, y=2, z=3, other=100)) == myfunc(1, 2, 3)
    assert 'unsupported operand type(s) for +' in str(error)


def test_get_defaults():
    assert get_defaults(defaults_fun) == {'default_var1': 10, 'default_var2': 20}


def test_cast_0_dim():
    to = np.array([1, 2, 3])
    casted_array = cast_0_dim(np.array(1), to=to)
    assert casted_array.shape == to.shape


def test_concat_solution():
    solutions = concat_solution((kamodofy(lambda x=np.array([1, 2, 3]): x ** 2) for i in range(1, 4)), 'y')
    expected = [1, 4, 9, np.nan, 1, 4, 9, np.nan, 1, 4, 9, np.nan, ]
    assert ((solutions['y'] == expected) | (np.isnan(solutions['y']) & np.isnan(expected))).all()


def test_get_unit_quantity():
    from kamodo import get_unit
    mykm = get_unit_quantity('mykm', 'km', scale_factor=2)
    mygm = get_unit_quantity('mygm', 'gram', scale_factor=4)
    assert str(mykm.convert_to(get_unit('m'))) == '2000.0*meter'
    assert str(mygm.convert_to(get_unit('kg'))) == '0.004*kilogram'


def test_substr_replace():
    mystr = "replace this string"
    mystr = substr_replace(mystr, [
        ('this', 'is'),
        ('replace', 'this'),
        ('string', 'replaced')
    ])
    assert mystr == 'this is replaced'


def test_beautify_latex():
    beautified = beautify_latex('LEFTx__plus1RIGHTminusLEFTacommabRIGHT')
    assert beautified == '\\left (x__+1\\right )-\\left (a,b\\right )'


def test_arg_to_latex():
    my_latex = arg_to_latex('x__iplus1')
    assert my_latex == 'x^{i+1}'


def test_simulate():
    def update_y(x):
        return x + 1

    def update_x(y):
        return y - 2

    state_funcs = OrderedDict([
        ('y', update_y),
        ('x', update_x),
        ('t', lambda t, dt: t + dt)
    ])

    simulation = simulate(state_funcs,
                          x=3,  # initial conditions
                          t=0,
                          dt=1,
                          steps=10)

    for state in simulation:
        pass

    assert state['x'] == -7
    assert state['y'] == -5
    assert state['t'] == 10


def test_simulate_error():
    def update_y(x):
        return x + '1'

    def update_x(y):
        return y - 2

    state_funcs = OrderedDict([
        ('y', update_y),
        ('x', update_x),
        ('t', lambda t, dt: t + dt)
    ])

    with pytest.raises(TypeError) as error:
        simulation = simulate(state_funcs,
                              x=3,
                              t=0,
                              dt=1,
                              steps=10)

        for state in simulation:
            pass

    assert "y:unsupported operand type(s) for +" in str(error)


def test_meta_units():
    @kamodofy(units='kg')
    def mass(x):
        return x

    assert mass.meta['units'] == 'kg'


def test_meta_data():
    @kamodofy(units='kg')
    def mass(x=list(range(5)), y=3):
        return [x_ + y for x_ in x]

    assert mass.data[-1] == 7


def test_meta_data_kwargs():
    @kamodofy(x=3, y=3)
    def e(x, y):
        return x + y

    assert e.data == 6

    @kamodofy(x=3)
    def f(x=5, y=3):
        return x + y

    assert f.data == 6

    @kamodofy()
    def g(x, y):
        return x + y

    assert g.data == None
    assert g.meta['units'] == ''

    @kamodofy(data=3)
    def h(x, y):
        return x + y

    assert h.data == 3


def test_repr_latex():
    @kamodofy(equation='$f(x) = x_i^2$')
    def f(x):
        return x

    assert f._repr_latex_() == 'f(x) = x_i^2'

    @kamodofy
    def g(x, y, z):
        return x
    print(g._repr_latex_())



def test_bibtex():
    bibtex = """
@phdthesis{phdthesis,
  author       = {Peter Joslin},
  title        = {The title of the work},
  school       = {The school of the thesis},
  year         = 1993,
  address      = {The address of the publisher},
  month        = 7,
  note         = {An optional note}
}"""

    @kamodofy(citation=bibtex)
    def h(x):
        '''This equation came out of my own brain'''
        return x

    assert '@phdthesis' in h.meta['citation']


def test_gridify():
    from kamodo import Kamodo
    kamodo = Kamodo(grids=grid_fun)
    assert kamodo.grids().shape == (10, 20, 30)



def test_pad_nan():
    array = np.array([[1, 2], [1, 2]])
    solution = pad_nan(array)
    expected = np.array([[1, 2], [1, 2], [np.nan, np.nan]])
    comparison = ((solution == expected) | (np.isnan(solution) & np.isnan(expected)))
    assert comparison.all()

    array = np.array([1, 2])
    solution = pad_nan(array)
    expected = np.array([1, 2, np.nan])
    comparison = ((solution == expected) | (np.isnan(solution) & np.isnan(expected)))
    assert comparison.all()

    with pytest.raises(NotImplementedError) as error:
        array = np.array([[[1], [2], [3]], [[1], [2], [3]]])
        pad_nan(array)

def test_pointlike():
    assert Kamodo(points=squeeze_point).points().shape == (100,)
    assert Kamodo(points=points).points().shape == (1, 100)
    assert Kamodo(points=squeeze_error_point).points().shape == (1, 100)


def test_solve():
    kamodo = Kamodo()
    seeds = np.array([1, 2])
    kamodo['fprime'] = solve(fprime, seeds, 'x', (0, 30), events=boundry)

    assert kamodo.fprime().shape == (2, 2)


def test_is_function():
    assert is_function(parse_expr('f(g(x))'))
    assert is_function(symbols('g', cls=UndefinedFunction))
    assert not is_function(symbols('x'))

def test_serialize_np():
    x = np.linspace(-5, 5, 12).reshape((3, 4))

    assert deserialize(serialize(x)).shape == (3, 4)
    assert deserialize(serialize(x))[0, 0] == -5


    x_json = json.dumps(x, default=serialize)
    x_ = json.loads(x_json, object_hook=deserialize)
    assert (x_ == x).all()

def test_serialize_pd():
    t = pd.date_range('Jan 1, 2021', 'Jan 11, 2021', freq='H')
    
    t_json = json.dumps(t, default=serialize)
    t_ = json.loads(t_json, object_hook=deserialize)
    assert (t_ == t).all()


    s = pd.Series(np.linspace(-5, 5, 10), index=list(range(10)))
    s_json = json.dumps(s, default=serialize)
    s_ = json.loads(s_json, object_hook=deserialize)
    assert (s_ == s).all()

    df = pd.DataFrame(np.linspace(-5, 5, 10), index=list(range(10)))
    df_json = json.dumps(df, default=serialize)
    df_ = json.loads(df_json, object_hook=deserialize)
    assert (df_ == df).all().all()

def test_serialize_generator():
    gen = (lambda x=b: x**2 for b in np.linspace(-5,5,12).reshape((4,3)))
    # need to reproduce the generator for comparison
    gen_duplicate = (lambda x=b: x**2 for b in np.linspace(-5,5,12).reshape((4,3)))
    gen_json = json.dumps(gen, default=serialize)
    gen_ = json.loads(gen_json, object_hook=deserialize)

    for f, f_ in zip(gen_, gen_duplicate):
        assert(f() == f_()).all()
        f_defaults = get_defaults(f)
        f__defaults = get_defaults(f_)
        for k, v in f_defaults.items():
            assert (v == f__defaults[k]).all()

def test_serialize_generator_pd():
    t = pd.date_range('Jan 1, 2021', 'Jan 11, 2021', freq='H')

    gent = (lambda t=t: (t - t0).total_seconds() for t0 in t[::10])

    for f in deserialize(serialize(gent)):
        assert len(f()) == len(t)
    assert f()[-1] == 0

def test_multiply_lambdagen():
    lamb = LambdaGenerator((lambda x=np.arange(i): x**2 for i in range(5)))
    lamb2 = LambdaGenerator((lambda x=np.arange(i): x**2 for i in range(5)))
    for r in lamb*lamb2:
        r()

def test_divide_lambdagen():
    lamb = LambdaGenerator(((lambda x=np.linspace(1, 5, 10): x**2) for i in range(5)))
    lamb2 = LambdaGenerator(((lambda x=np.linspace(1, 5, 10): x**2) for i in range(5)))
    for r in lamb/lamb2:
        r()

def test_get_args():
    def f(x, y, z):
        return x+y+z

    assert get_args(f) == ('x', 'y', 'z')

def test_curry():
    @curry
    def arimean(*args):
        return sum(args) / len(args)

    assert arimean(-2, -1, 0, 1, 2)() == 0
    assert arimean(-2)(-1)(0)(1)(2)() == 0

def test_partial_decorator():
    @partial(z=1, verbose=True)
    def f(x, y=2, z=5):
        """simple function"""
        return x + y + z
    assert f(2,3) == 2+3+1
    

def test_partial_order():
    @partial(y=1, verbose=True)
    def f(x, y=2, z=5):
        """simple function"""
        return x + y + z
    assert f(3,4) == 3+1+4

def test_partial_no_kwargs():
    @partial(verbose=True)
    def f(x, y=2, z=5):
        """simple function"""
        return x + y + z
    assert f(3,4) == 3+4+5


def test_partial_inline():
    def f(x, y=2, z=5):
        """simple function"""
        return x + y + z
    g = partial(f, y=5, verbose=True)
    assert(g(3,4) == 3+5+4)

def test_partial_no_defaults():
    @partial(x=1, y=2, z=3, verbose=True)
    def f(x, y, z):
        return x + y + z
    assert f() == 1 + 2 + 3

def test_partial_missing_defaults():
    @partial(x=1, y=2, verbose=True)
    def f(x,y,z):
        return x+y+z
    assert f(3) == 1 + 2 + 3

def test_partial_required_args():
    """need to make z a required argument"""
    @partial(x=1, y=2, verbose=True)
    def f(x,y,z):
        return x+y+z
    try:
        f()
    except TypeError as m:
        assert 'missing' in str(m)

def test_partial_docs():
    @partial(y=2)
    def f(x, y=3, z=4):
        """my docs"""
        return x + y + z
    
    assert 'my docs' in f.__doc__
    assert 'f(x, z=4) = f(x, y=2, z)' in f.__doc__

    @partial(y=2, z=1)
    def f(x, y=3, z=4):
        """my docs"""
        return x + y + z

    assert 'Calling f(x, y, z) for fixed y, z' in f.__doc__


    


# -*- coding: utf-8 -*-
"""
Copyright © 2017 United States Government as represented by the Administrator, National Aeronautics and Space Administration.  
No Copyright is claimed in the United States under Title 17, U.S. Code.  All Other Rights Reserved.
"""
import plotly.graph_objs as go
import numpy as np
from .util import arg_to_latex, beautify_latex, cast_0_dim, get_defaults
from plotly import figure_factory as ff
import pandas as pd
from collections import defaultdict
from .util import get_bbox

def scatter_plot(result, titles, verbose = False, **kwargs):
    """Generates a 3d scatter plot


    result: a dictionary of parameters
    titles: a dictionary of titles

    returns:
        [trace], chart_type, layout
    """
    if verbose:
        print('3-d scatter plot')

    # get the name of this variable
    variable = titles['variable']

    # get the first parameter from results
    val0 = list(result.values())[0] 

    # variable should be in result dictionary
    if result[variable].shape[0] == val0.shape[0]:
        if verbose:
            print('\t 3-d output', result[variable].shape, val0.shape)
            print('\t 3d scatter plot')

        if type(val0) == pd.DataFrame:
            p_x = val0.values[:, 0].tolist()
            p_y = val0.values[:, 1].tolist()
            p_z = val0.values[:, 2].tolist()
        else:
            p_x = val0[:, 0].tolist()
            p_y = val0[:, 1].tolist()
            p_z = val0[:, 2].tolist()

        trace = go.Scatter3d(x=p_x,
                             y=p_y,
                             z=p_z,
                             mode='markers',
                             marker=dict(color=result[variable],
                                         colorscale='Viridis'),
                             )

        layout = go.Layout(
            title=titles['title'],
            scene=dict(
                xaxis=dict(title='x'),
                yaxis=dict(title='y'),
                zaxis=dict(title='z'),
                ))
        chart_type = '3d-scatter'
    else:
        print(result[variable].shape, val0.shape)
        raise NotImplementedError('Not implemented yet')

    return [trace], chart_type, layout



def line_plot(result, titles, verbose=False, **kwargs):
    '''N-d line plot f(t)'''
    if verbose:
        print('N-d line plot f(t)')
    f = result[titles['variable']]
    t_name, t = list(result.items())[0]

    # one input, one output
    if len(result) == 2:
        # if the output is a pandas dataframe
        if type(f) == pd.DataFrame:
            # if the output is a pandas multiindex
            # this code is only used for an experimental format for fieldlines
            if type(f.index) == pd.MultiIndex:
                assert len(t) == len(f)
                l = []
                t_start = 0
                t_ = []
                for seed, locs in f.groupby(level=0):
                    locs = locs.append(pd.Series(), ignore_index=True)
                    l.append(locs)
                    t_.extend(t[t_start:t_start+len(locs)].tolist() + [np.nan])
                f = pd.concat(l)
                t = t_

        # if the output is 1-dimensional, assume a basic x vs y plot
        if len(f.shape) == 1:
            if verbose:
                print('\t1-d output', f.shape)
            assert f.shape == t.shape
            y = f
            layout = go.Layout(
                title=titles['title'],
                xaxis=dict(title='${}$'.format(t_name)),
                yaxis=dict(title=titles['title_lhs']))
            trace = go.Scatter(x=t, y=y, name=titles['title_short'])
            chart_type = 'line'

        # if the output is 2-dimentional, assume a 2-d parametric line plot
        elif f.shape[1] == 2:
            if verbose:
                print('\t2-d output', f.shape)
            # t_name, t = result.items()[0]
            text = ["{}:{}".format(t_name, v) for v in t]
            if type(f) == pd.DataFrame:
                x = f.values[:, 0]
                y = f.values[:, 1]
                trace = go.Scatter(x=x, y=y, text=text)
                layout = go.Layout(title=titles['title'],
                                   xaxis=dict(title=f.columns[0] + titles['units']),
                                   yaxis=dict(title=f.columns[1] + titles['units']))
            else:
                x = f[:, 0]
                y = f[:, 1]
                trace = go.Scatter(x=x, y=y, text=text)
                layout = go.Layout(title=titles['title'],
                                   xaxis=dict(title='x' + titles['units']),
                                   yaxis=dict(title='y' + titles['units']))


            chart_type = '2d-line'

        # if the output shape is 3-dimensional, make a 3-d parametric line plot
        elif f.shape[1] == 3:
            if verbose:
                print('\t3-d output', f.shape)
            # ToDo: for pandas, use the names of the columns as the coordinates
            if type(f) == pd.DataFrame:
                x = f.values[:, 0]
                y = f.values[:, 1]
                z = f.values[:, 2]
                layout = go.Layout(
                    title=titles['title'],
                    scene=dict(
                        xaxis=dict(title=f.columns[0] + titles['units']),
                        yaxis=dict(title=f.columns[1] + titles['units']),
                        zaxis=dict(title=f.columns[2] + titles['units']),
                        ))
            else:
                x = f[:, 0]
                y = f[:, 1]
                z = f[:, 2]
                layout = go.Layout(
                    title=titles['title'],
                    scene=dict(
                        xaxis=dict(title='x' + titles['units']),
                        yaxis=dict(title='y' + titles['units']),
                        zaxis=dict(title='z' + titles['units']),
                        ))
            # t_name, t = result.items()[0]
            text = ["{}:{}".format(t_name, v) for v in t]

            trace = go.Scatter3d(x=x, y=y, z=z, text=text, mode='lines')
            chart_type = '3d-line'

    # two inputs, one output
    # here we intended to make a colored line plot in 2 dimensions
    # however, there may be a bug in plotly that prevented this from working properly
    elif len(result) == 3:
        arg0, val0 = t_name, t
        arg1, val1 = list(result.items())[1]

        assert f.shape == t.shape

        text = ["{}:{}".format(titles['variable'], v) for v in f]

        trace = go.Scatter(
            x=val0,
            y=val1,
            text=text,
            # marker=dict(
            #   color=result[titles['variable']],
            #   colorscale='Viridis',
            #   colorbar=dict(title=titles['title_short'])),
            #  since plotly doesn't support autocolor for 2d lines yet
            line=dict(
                # color=result[titles['variable']],
                color='black', # need to map this to a colorscale
                # colorscale='Viridis',
                ),
            mode='lines')

        layout = dict(
            title=titles['title'],
            xaxis_title=arg0,
            yaxis_title=arg1,
            scene=dict(
                xaxis=dict(title=arg0),
                yaxis=dict(title=arg1),
                ))
        chart_type = '2d-line'

    # 3 inputs, one output
    # assume a 3d parametric line with color
    elif len(result) == 4:
        arg0, val0 = t_name, t
        arg1, val1 = list(result.items())[1]
        arg2, val2 = list(result.items())[2]

        assert(f.shape == t.shape)

        text = ["{}:{}".format(titles['variable'], v) for v in f]

        trace = go.Scatter3d(
            x=val0,
            y=val1,
            z=val2,
            text=text,
            # marker=dict(
            #   color=result[titles['variable']],
            #   colorscale='Viridis',
            #   colorbar=dict(title=titles['title_short'])),
            line=dict(
                color=result[titles['variable']],
                colorscale='Viridis'),
            mode='lines')
        layout = go.Layout(
            title=titles['title'],
            scene=dict(
                xaxis=dict(title=arg0),
                yaxis=dict(title=arg1),
                zaxis=dict(title=arg2),))
        chart_type = '4d-line'
    else:
        raise NotImplementedError('shape not supported: {}'.format(f.shape))

    return [trace], chart_type, layout


def vector_plot(result, titles, verbose = False, **kwargs):
    variable = titles['variable']
    val0 = list(result.values())[0]

    # check if this is a 2d vector plot: input and output should have shape 2d
    if (result[variable].shape == val0.shape) & (val0.shape[1] == 2):
        if verbose:
            print('\t 2-d output', result[variable].shape)
        u = result[variable][:, 0]
        v = result[variable][:, 1]
        x = val0[:, 0]
        y = val0[:, 1]

        # plotly's quiver plot has its own defaults
        quiver_defaults = get_defaults(ff.create_quiver)
        for k, v_ in kwargs.items():
            if k in quiver_defaults:
                quiver_defaults[k] = v_
        if verbose:
            print(quiver_defaults)
        try:
            trace = ff.create_quiver(
                x, y, u, v,
                name='quiver',
                line=dict(width=1),
                **quiver_defaults)['data'][0] # extracts quiver trace
        except:
            print('problem with quiver, u,v,x,y shapes:')
            for v_ in [u, v, x, y]:
                print(v_.shape)
            raise

        layout = go.Layout(title=titles['title'],
                           xaxis=dict(title='x'),
                           yaxis=dict(title='y'))
        chart_type = '2d-vector'

    # check if this is a 3d vector plot: input and output should have shape 3d
    elif (result[variable].shape == val0.shape) & (val0.shape[1] == 3):
        if verbose:
            print('\t 3-d output', result[variable].shape, val0.shape)
            print('\t 3d vector plot')
        if type(result[variable]) == pd.DataFrame:
            u = result[variable].values[:, 0].tolist()
            v = result[variable].values[:, 1].tolist()
            w = result[variable].values[:, 2].tolist()
        else:
            u = result[variable][:, 0].tolist()
            v = result[variable][:, 1].tolist()
            w = result[variable][:, 2].tolist()
        if type(val0) == pd.DataFrame:
            x = val0.values[:, 0].tolist()
            y = val0.values[:, 1].tolist()
            z = val0.values[:, 2].tolist()
        else:
            x = val0[:, 0].tolist()
            y = val0[:, 1].tolist()
            z = val0[:, 2].tolist()

        # normalize each vector to get the length
        norms = np.linalg.norm(result[variable], axis=1)
        trace = go.Cone(
            x=x, y=y, z=z, u=u, v=v, w=w,
            hoverinfo='x+y+z+u+v+w+norm',
            colorscale='Reds',
            cmin=0, cmax=norms.max(),
            )

        layout = go.Layout(
            title=titles['title'],
            scene=dict(
                xaxis=dict(title='x'),
                yaxis=dict(title='y'),
                zaxis=dict(title='z'),
                ))
        chart_type = '3d-vector'
    else:
        raise NotImplementedError('plot type not supported yet')

    return [trace], chart_type, layout

def contour_plot(result, titles, indexing, verbose = False, **kwargs):
    variable = titles['variable']
    if verbose:
        print('\t2-d output', result[variable].shape)
    z = result[variable]
    title = titles['title']
    arg0, val0 = list(result.items())[0]
    arg1, val1 = list(result.items())[1]

    if verbose:
        print('2 inputs: {} {}, {} {}'.format(arg0, val0.shape, arg1, val1.shape))
    traces = []
    if len(val0.shape) == 1:
        if verbose:
            print('\t\t1-d args', end=' ')
        if result[variable].shape[0] == val0.shape[0]:
            if indexing == 'ij':
                if verbose:
                    print('{} indexing'.format(indexing))
                trace = go.Contour(x = val0,
                                   y = val1,
                                   z = z.T)
            else:
                if verbose:
                    print('{} indexing'.format(indexing))
                trace = go.Contour(x = val0,
                                   y = val1,
                                   z = z)

        else:
            if verbose:
                print('xy indexing')
            trace = go.Contour(x = val0,
                               y = val1,
                               z = z)
        layout = go.Layout(
            title = title,
            xaxis = dict(title = '${}$'.format(arg0)),
            yaxis = dict(title = '${}$'.format(arg1)))
        traces.append(trace)
        chart_type = '2d-grid'
    else:
        if verbose:
            print('\t\t2-d args', val0.shape, val1.shape)
        assert val0.shape == val1.shape
        assert val0.shape == z.shape
        xaxis = dict(title = beautify_latex('${}$'.format(arg_to_latex(arg0))))
        yaxis = dict(title = beautify_latex('${}$'.format(arg_to_latex(arg1))))
        carpet_traces, layout = carpet_plot(result, title, xaxis, yaxis)
        traces.extend(carpet_traces)
        chart_type = '2d-skew'

    return traces, chart_type, layout

def carpet_plot(results, title, xaxis, yaxis, indexing = 'xy', **kwargs):
    '''Assumes ordered dict where values have the same shape'''
    # print title, xaxis, yaxis
    arg0, val0 = list(results.items())[0]
    arg1, val1 = list(results.items())[1]
    arg2, val2 = list(results.items())[2]
    
    a = list(range(val0.shape[0]))
    b = list(range(val0.shape[1]))

    aa, bb = np.meshgrid(b,a, indexing = indexing)

    trace1 = go.Contourcarpet(
        a = aa.ravel(),
        b = bb.ravel(),
        z = val2.ravel(),
    )

    trace2 = go.Carpet(
        a = aa.ravel(),
        b = bb.ravel(),
        x = val0.ravel(),
        y = val1.ravel(),
        aaxis = dict(
            tickprefix = ''.format(arg0),
            smoothing = 0,
            minorgridcount = 0,
            type = 'linear',
            showgrid = False,
            nticks = 5,
            dtick=0,
            tickmode='linear',
            showticklabels='none',
        ),
        baxis = dict(
            tickprefix = ''.format(arg1),
            smoothing = 0,
            minorgridcount = 0,
            type = 'linear',
            showgrid = False,
            nticks = 5,
            dtick=0,
            tickmode='linear',
            showticklabels='none',
        )
    )
    traces = [trace1, trace2]
    layout = go.Layout(
        title = title, 
        xaxis = xaxis,
        yaxis =  yaxis)
    return traces, layout

def plane(result, titles, indexing = 'xy', verbose = False, **kwargs):
    variable = titles['variable']
    arg0, val0 = list(result.items())[0]
    arg1, val1 = list(result.items())[1]
    arg2, val2 = list(result.items())[2]
    traces = []
    if verbose:
        print('\t3-d output', result[variable].shape)
    surfacecolor = result[variable].squeeze()
    create_meshgrid = True
    for x in val0, val1, val2:
        if len(get_arg_shapes(x)[0]) != 1:
            create_meshgrid = False
            ones = np.ones_like(x)
    if create_meshgrid:
        xx, yy, zz = np.meshgrid(val0, val1, val2)
    else:
        xx, yy, zz = [v*ones for v in [val0, val1, val2]]
    traces.append(go.Surface(
        x = xx.squeeze(),
        y = yy.squeeze(),
        z = zz.squeeze(),
        surfacecolor = surfacecolor))
    layout = go.Layout(
        title = titles['title'],
        scene = dict(
            xaxis = dict(title = arg0),
            yaxis = dict(title = arg1),
            zaxis = dict(title = arg2),
        )
    )
    chart_type = '3d-plane'

    return traces, chart_type, layout

def surface(result, titles, verbose = False, **kwargs):
    variable = titles['variable']
    title = titles['title']
    arg0, val0 = list(result.items())[0]
    arg1, val1 = list(result.items())[1]
    arg2, val2 = list(result.items())[2]

    if verbose:
        print('\t2-d output', result[variable].shape)

    traces = []

    if len(result[variable].shape) == 2:
        surfacecolor = result[variable]
        xx = cast_0_dim(val0, result[variable])
        yy = cast_0_dim(val1, result[variable])
        zz = cast_0_dim(val2, result[variable])
        traces.append(go.Surface(
            x = xx,
            y = yy,
            z = zz,
            surfacecolor = surfacecolor))
        layout = go.Layout(
            title = title,
            scene = dict(
                xaxis = dict(title = arg0),
                yaxis = dict(title = arg1),
                zaxis = dict(title = arg2),))
        chart_type = '3d-surface-scalar'

    elif len(result[variable].shape) == 1:
        if verbose:
            print('\tscalar output', result[variable].shape)

        elif result[variable].shape[0] == 1:
            traces.append(go.Surface(
                x = val0, 
                y = val1, 
                z = val2, 
                ))
            layout = go.Layout(
                title = title,
                scene = dict(
                    xaxis = dict(title = arg0),
                    yaxis = dict(title = arg1),
                    zaxis = dict(title = arg2),
                    ))
            chart_type = '3d-surface'

    elif len(result[variable]) == 1: #scalar or color value?
        if verbose:
            print('\tsingle valued output', result[variable].shape)
        surfacecolor = result[variable]
        arg0, val0 = list(result.items())[0]
        arg1, val1 = list(result.items())[1]
        arg2, val2 = list(result.items())[2]
        xx = val0
        yy = val1
        zz = val2
        traces.append(go.Surface(
            x = xx,
            y = yy,
            z = zz,
            ))
        layout = go.Layout(
            title = title,
            scene = dict(
                xaxis = dict(title = arg0),
                yaxis = dict(title = arg1),
                zaxis = dict(title = arg2),
            )
        )

    return traces, chart_type, layout

def tri_surface_plot(result, titles, verbose = False, **kwargs):
    # triangulated surface 
    variable = titles['variable']
    results = list(result.items())
    arg0, val0 = results[0]
    arg1, val1 = results[1]
    arg2, val2 = results[2]
    
    trace = go.Mesh3d(
        x=val0,
        y=val1,
        z=val2,
        colorbar_title = variable,
        colorscale=[[0, 'gold'],
                    [0.5, 'mediumturquoise'],
                    [1, 'magenta']],
        # Intensity of each vertex, which will be interpolated and color-coded
        intensity = val2,
        # i, j and k give the vertices of triangles
        i=result[variable][:,0],
        j=result[variable][:,1],
        k=result[variable][:,2],
        name=variable,
        showscale=True)
    
    layout = go.Layout(
        title = titles['title'],
        scene = dict(
            xaxis = dict(title = arg0),
            yaxis = dict(title = arg1),
            zaxis = dict(title = arg2),))
    return [trace], '3d-surface', layout

def image(result, titles, verbose=False, **kwargs):

    variable = titles['variable']
    if verbose:
        print('\t2-d image', result[variable].shape)
    z = result[variable]
    title = titles['title']
    arg0, val0 = list(result.items())[0]
    arg1, val1 = list(result.items())[1]

    if verbose:
        print('2 inputs: {} {}, {} {}'.format(arg0, val0.shape, arg1, val1.shape))

    trace = go.Image(z=z)
    layout = go.Layout(
        title = title,
        xaxis = dict(title = '${}$'.format(arg0)),
        yaxis = dict(title = '${}$'.format(arg1)))

    return [trace], '2d-image', layout


# {output.shape : {(input1.shape, input2.shape) : {'name':plot_name, 'func': plot_func}}}

plot_dict = {
    (1,)    :   {(('N','M'), ('N','M'), ('N','M')): {'name': '3d-parametric', 'func': surface}},
    ('N',)  :   {
        (('N',),) : {'name': '1d-line', 'func': line_plot},
        (('N',), ('N',)): {'name': '2d-line-scalar', 'func': line_plot},
        (('N',), ('N',),('N',)): {'name': '3d-line-scalar', 'func': line_plot},
        (('N', 3),): {'name': '3d scatter', 'func': scatter_plot},},
    ('N', 2) :   {
        (('N',),) :{'name': '2d-line', 'func': line_plot},
        (('N', 2),):{'name': '2d-vector', 'func': vector_plot},
    },
    ('N', 3) :   {
        (('N',),) :{'name': '3d-line', 'func': line_plot},
        (('N', 3),):{'name': '3d-vector', 'func': vector_plot},
        (('M',), ('M',), ('M',)):{'name': '3d-tri-surface', 'func': tri_surface_plot},
    },
    ('N', 'N'): {
        (('N',), ('N',)) :{'name': '2d-contour', 'func': contour_plot},
    },
    ('N', 'M'):  {
        (('N',), ('M',)) :{'name': '2d-contour', 'func': contour_plot},
        (('M',), ('N',)) :{'name': '2d-contour', 'func': contour_plot},
        (('N', 'M'), ('N','M')):{'name': '2d-contour-skew', 'func': contour_plot},
        (('N', 'M'), ('N','M'), ('N','M')): {'name': '3d-parametric-scalar', 'func': surface},
        ((1,), ('N', 'M'),('N','M')):{'name': '3d-plane', 'func': plane},
        (('N', 'M'), (1,),('N','M')):{'name': '3d-plane', 'func': plane},
        (('N', 'M'), ('N','M'),(1,)):{'name': '3d-plane', 'func': plane},
    },
    ('N', 1, 'M'): {
        (('N',), (1,), ('M',)):{'name': '3d-plane', 'func': plane},
        ((1,), ('N',), ('M',)):{'name': '3d-plane', 'func': plane},
    },
    (1, 'N', 'M'): {
        (('N',), (1,), ('M',)):{'name': '3d-plane', 'func': plane},
    },
    ('N', 'M', 1): {
        ((1,), ('N',), ('M',)):{'name': '3d-plane', 'func': plane},
        (('N',), (1,), ('M',)):{'name': '3d-plane', 'func': plane},
        (('N',), ('M',), (1,)):{'name': '3d-plane', 'func': plane},
        (('M',), ('N',), (1,)):{'name': '3d-plane', 'func': plane},
    },
    ('N', 'M', 3): {
        (('N',), ('M',)):{'name': 'image', 'func': image}
    },
}

# keep values below size threshold
size_threshold = 3
# +
sizes_available = np.array(['N', 'M', 'L', 'O', 'P', 'Q', 'R', 'S', 'T'], dtype=object)

def flatten_shapes(shapes):
    return [item for sublist in shapes for item in sublist]

def unordered_unique(a):
    """return the indices and values of the array in their original order"""
    indices = []
    result = []
    j = 0
    for i, _ in enumerate(a):
        if _ not in result:
            result.append(_)
        indices.append(result.index(_))
    return result, indices

def symbolic_shape(*shapes, sizes_available=sizes_available):
    """Convert input shapes to symbolic shapes

    Allow values of 1 to pass through
    Results should match input structure
    """
    flattened = np.array(flatten_shapes(shapes), dtype=object)

    swappable_indices = np.argwhere(flattened > size_threshold).T[0]
    swappable_vals = flattened[swappable_indices]

    _, unique = unordered_unique(swappable_vals)

    # put results back in the original orray
    flattened[swappable_indices] = sizes_available[unique]

    # result comes out flattened, restructure to match input
    shape_index = 0
    results = []
    for shape in shapes:
        result_shape = []
        for s_ in shape:
            try:
                result_shape.append(flattened[shape_index])
            except IndexError:
                print('input shapes:', shapes)
                raise
            shape_index += 1
        results.append(tuple(result_shape))
    return tuple(results)
# -

def get_arg_shapes(*args):
    shapes = []
    for a in args:
        if type(a) == np.ndarray:
            shape = a.shape
        elif type(a) == pd.DataFrame:
            shape = a.values.shape
        else:
            try:
                shape = (len(a),)
            except:
                shape = (1,)
        shapes.append(shape)
    return shapes


# +
def func_mod_name(f):
    return '{}.{}'.format(f.__module__, f.__name__)

def get_plot_types_df():
    """pack the plot types into a dataframe"""
    plot_types = dict()
    for out_shape, v in list(plot_dict.items()):
        for arg_shapes, v_ in list(v.items()):
            plot_key = out_shape, arg_shapes
            if plot_key in plot_types:
                raise KeyError('plot_key already present {}'.format(plot_key))
            plot_types[plot_key] = [v_['name'], func_mod_name(v_['func'])]
    
    plot_types = pd.DataFrame(plot_types).T
    
    plot_types.index.set_names(['out_shape', 'arg_shapes'], inplace = True)
    
    plot_types.columns = ['plot_type', 'function']
    return plot_types
    
plot_types = get_plot_types_df()


def get_ranges(figures):
    axes_min = defaultdict(list)
    axes_max = defaultdict(list)

    axis_names = 'xaxis', 'yaxis', 'zaxis'


    for fig in figures:
        ranges = get_bbox(fig)
        for i, val in enumerate(ranges):
            axname = axis_names[i//2]
            if i%2 == 0:
                axes_min[axname].append(val)
            else:
                axes_max[axname].append(val)
    for axis in axes_min:
        axes_min[axis] = min(axes_min[axis])
        axes_max[axis] = max(axes_max[axis])


    axes = dict()
    if len(axes_min) == 2: # 2d
        for axis in 'xaxis', 'yaxis':
            axes[axis] = dict(autorange=False,
                range=(axes_min[axis], axes_max[axis]))
    else: # 3d
        axes = dict(scene=dict(aspectmode='manual'))

        for axis in 'xaxis', 'yaxis', 'zaxis':
            axes['scene'][axis] = dict(autorange=False,
                range=(axes_min[axis], axes_max[axis]))
        aspectratio = dict()
        for _ in 'xyz':
            min_, max_ = axes['scene'][_+'axis']['range']
            aspectratio[_] = max_ - min_
        axes['scene']['aspectratio'] = aspectratio
        axes['scene']['camera'] = dict(eye={_:axes['scene'][_ + 'axis']['range'][1] for _ in 'xyz'})

    return axes





import os
from os import path

import numpy as np
import pandas as pd
from kamodo import kamodofy, Kamodo
from sympy.core.function import UndefinedFunction
from kamodo import get_defaults, getfullargspec
from kamodo import serialize, deserialize


import flask
from flask import Flask, render_template, jsonify
from flask_cors import CORS, cross_origin
from flask_restful import reqparse, abort, Api, Resource
from flask import request

from io import StringIO
import json

import hydra
from hydra.experimental import compose
from omegaconf import OmegaConf

import plotly.graph_objs as go

try:
    hydra.experimental.initialize(strict=False)
except:
    pass

from kamodo.util import NumpyArrayEncoder

from collections import defaultdict
from kamodo import from_kamodo


import logging
# log = logging.getLogger('werkzeug')
# log.setLevel(logging.ERROR)


app = Flask(__name__)

CORS(app) # enable Cross Origin Resource Sharing (CORS), making cross-origin AJAX possible.


def config_override(cfg):
    """Overrides with user-supplied configuration

    kamodo will override its configuration using
    kamodo.yaml if it is in the current working directory
    or users can set an override config:
        config_override=path/to/myconfig.yaml
    """
    if cfg.config_override is not None:
        override_path = hydra.utils.to_absolute_path(cfg.config_override)
        if path.exists(override_path):
            override_conf = OmegaConf.load(override_path)
            # merge overrides first input with second
            cfg = OmegaConf.merge(cfg, override_conf)
        else:
            print('no such path: {}\n ..ignoring'.format(override_path))
    return cfg


def get_api(app):
    """get restful api for this app"""
    api = Api(app)
    return api

# models registered at startup
global_models = {}

# models registered by users, providing access to global models
user_models = {}

def main():
    """main entrypoint"""

    cfg = compose('conf/kamodo.yaml')

    cli_conf = OmegaConf.from_cli()
    cfg = OmegaConf.merge(cfg, cli_conf)

    extra_files = []
    config_override_ = None
    if cfg.config_override is not None:
        override_path = "{}/{}".format(os.getcwd(), cfg.config_override)
        if path.exists(override_path):
            app.logger.info("found {}".format(override_path))
            config_override_ = OmegaConf.load(override_path)
            app.logger.info(config_override)
            extra_files.append(override_path)
        else:
            if cfg.verbose > 0:
                app.logger.info("could not get override: {}".format(override_path))
                print("could not get override: {}".format(override_path))

        if config_override_ is not None:
            cfg = OmegaConf.merge(cfg, config_override_)

            # make sure models is replaced
            if 'models' in config_override_:
                cfg['models'] = config_override_.models


    if cfg.verbose > 0:
        app.logger.info(cfg.pretty())


    for model_name, model_conf in cfg.models.items():
        try:
            # this might fail
            model_ = hydra.utils.instantiate(model_conf)
            global_models[model_name] = model_
        except Exception as m:
            print(m)
            if cfg.verbose:
                print('could not start {}'.format(model_name))
            pass

    api = Api(app)


    api.add_resource(get_global_models_resource(), '/api', '/api/')


    for model_name, model_ in global_models.items():
        register_endpoints(api, model_name, model_, '/api')


    global user_models
    user_model_default = hydra.utils.instantiate(cfg.user_model)
    user_models = defaultdict(lambda: from_kamodo(user_model_default))

    api.add_resource(get_user_models_resource(), '/kamodo/api', '/kamodo/api/')


    @app.route('/kamodo/api/<user_model_name>', methods=['GET', 'POST'])
    @app.route('/kamodo/api/<user_model_name>/', methods=['GET', 'POST'])
    @app.route('/kamodo/api/<user_model_name>/<user_func>', methods=['GET', 'POST', "DELETE"])
    def kamodo_model_func(user_model_name, user_func=None):
        user_model = user_models[user_model_name]
        print(user_model_name, user_func)
        print(user_model)

        if request.method == 'POST':
            func_key = request.form['signature']
            if 'expression' in request.form:
                func_expr = request.form['expression']
                if func_key in user_model:
                    del user_model[func_key]
                user_model[func_key] = func_expr
            elif 'model' in request.form:
                global_model_name = request.form['model']
                global_model = global_models[global_model_name]
                if 'model_func' in request.form:
                    global_model_func_name = request.form['model_func']
                    global_model_func = global_model[global_model_func_name]
                    user_model[func_key] = global_model_func

        if user_func is not None:
            parser = reqparse.RequestParser()
            if request.method == 'GET':
                func = user_model[user_func]
                defaults = get_defaults(func)
                for arg in getfullargspec(func).args:
                    if arg in defaults:
                        parser.add_argument(arg, type=str)
                    else:
                        parser.add_argument(arg, type=str, required=True)

                args_ = parser.parse_args(strict=True)
                args = dict()

                for argname, val_ in args_.items():
                    args[argname] = json.loads(val_, object_hook=deserialize)
                    if isinstance(args[argname], str):
                        args[argname] = json.loads(val_, object_hook=deserialize)
                    if isinstance(args[argname], list):
                        args[argname] = np.array(args[argname])
                print('{} {} passed {}'.format(
                    model_name,
                    user_func,
                    ' '.join(['{}:{}'.format(arg, type(arg_value)) for arg, arg_value in args.items()])))
                try:
                    result = func(**args)
                    print('{} {} function returned {}'.format(
                        model_name, user_func, type(result)))
                    return serialize(result)
                except:
                    return json.dumps(None)
            if request.method == "DELETE":
                if user_func in user_model:
                    del user_model[user_func]

        return get_model_details(user_model)
        # return dict(user_func = user_func, user_model_name = user_model_name)

    @app.route('/kamodo/api/<user_model_name>/<user_func>/defaults', methods=['GET'])
    @app.route('/kamodo/api/<user_model_name>/<user_func>/defaults/', methods=['GET'])
    def kamodo_model_func_defaults(user_model_name, user_func):
        user_model = user_models[user_model_name]
        user_func = user_model[user_func]
        function_defaults = get_defaults(user_func)

        for arg in getfullargspec(user_func).args:
            if arg in function_defaults:
                pass
            else:
                function_defaults[arg] = None
        try:
            function_defaults_ = json.dumps(function_defaults, default=serialize)
        except:
            print('problem with {}.{}'.format(model_name, user_func))
            raise
        return jsonify(function_defaults_)

    @app.route('/kamodo/api/<user_model_name>/<user_func>/data', methods=['GET'])
    @app.route('/kamodo/api/<user_model_name>/<user_func>/data/', methods=['GET'])
    def kamodo_model_func_data(user_model_name, user_func):
        user_model = user_models[user_model_name]
        func = user_model[user_func]
        # assume function is kamodofied
        return jsonify(json.dumps(func.data, default=serialize))
        # function_defaults = get_defaults(func)
        # func_data = func(**function_defaults)
        # return jsonify(serialize(func_data))

    @app.route('/kamodo/api/<user_model_name>/<user_func>/plot', methods=['GET'])
    @app.route('/kamodo/api/<user_model_name>/<user_func>/plot/', methods=['GET'])
    def kamodo_model_func_plot(user_model_name, user_func):
        user_model = user_models[user_model_name]
        func = user_model[user_func]
        defaults = get_defaults(func)

        parser = reqparse.RequestParser()

        for arg in getfullargspec(func).args:
            if arg in defaults:
                parser.add_argument(arg, type=str)
            else:
                parser.add_argument(arg, type=str, required=True)


        app.logger.info('function plot resource called')
        args_ = parser.parse_args(strict=True)
        args = dict(indexing='ij') # needed for gridded data

        for argname, val_ in args_.items():
            if val_ is not None:
                args[argname] = pd.read_json(StringIO(val_), typ='series')
        figure = user_model.figure(variable=user_func, **args)
        return go.Figure(figure).to_json()



    try:
        app.run(host=cfg.flask.host, port=cfg.flask.port)
    except OSError as m:
        print('cannot start with configuration', cfg.flask)
        raise




def get_model_details(model):
    detail = model.detail().astype(str)
    detail_dict = detail.to_dict(
                    # default_handler=str,
                    # indent =4,
                    orient='index',
                    )
    return detail_dict

def get_global_models_resource():
    '''registers signatures of multiple models'''
    class GlobalModels(Resource):
        def get(self):
            details = dict()
            for model_name, model_ in global_models.items():
                details[model_name] = get_model_details(model_)
            return details
    return GlobalModels

def get_user_models_resource():
    '''registers signatures of multiple models'''
    class UserModels(Resource):
        def get(self):
            details = dict()
            for model_name, model_ in user_models.items():
                details[model_name] = get_model_details(model_)
            return details
    return UserModels

def register_func_endpoints(api, model_name, model_, base_name, var_symbol):
    var_label = str(type(var_symbol))

    # /api/mymodel/myfunc
    func_resource_endpoint = '{}/{}/{}'.format(base_name, model_name, var_label)
    print('registering ' + func_resource_endpoint)
    try:
        api.add_resource(
            get_func_resource(model_name, model_, var_symbol),
            func_resource_endpoint,
            endpoint='/'.join([model_name, var_label]))

        # /api/mymodel/myfunc/defaults
        defaults_resource_endpoint = '{}/{}/{}/{}'.format(
            base_name, model_name, var_label, 'defaults')
        print('registering ' + defaults_resource_endpoint)
        api.add_resource(
            get_defaults_resource(model_name, model_, var_symbol),
            defaults_resource_endpoint,
            endpoint='/'.join([model_name, var_label, 'defaults']))

        # /api/mymodel/myfunc/data
        data_resource_endpoint = '{}/{}/{}/{}'.format(
            base_name, model_name, var_label, 'data')
        print('registering ' + data_resource_endpoint)
        api.add_resource(
            get_data_resource(model_name, model_, var_symbol),
            data_resource_endpoint,
            endpoint='/'.join([model_name, var_label, 'data']))

        # /api/mymodel/myfunc/plot
        func_plot_resource_endpoint = '{}/{}/{}/{}'.format(
            base_name, model_name, var_label, 'plot')
        print('registering ' + func_plot_resource_endpoint)
        api.add_resource(
            get_func_plot_resource(model_, var_symbol),
            func_plot_resource_endpoint,
            endpoint='/'.join([model_name, var_label, 'plot']))

        # /api/mymodel/myfunc/doc
        func_doc_resource_endpoint = '{}/{}/{}/{}'.format(
            base_name, model_name, var_label, 'doc')
        print('registering ' + func_doc_resource_endpoint)
        api.add_resource(
            get_func_doc_resource(model_, var_symbol),
            func_doc_resource_endpoint,
            endpoint='/'.join([model_name, var_label, 'doc']))

    except ValueError as m:
        print('warning: {}'.format(m))
        pass



def register_endpoints(api, model_name, model_, base_name, register_base=True):
    '''register all endpoints for this model'''
    print('registering endpoints for {}/{}'.format(base_name, model_name))
    logging.info('registering endpoints for {}/{}'.format(base_name, model_name))

    if register_base:
        model_resource_endpoint = '{}/{}'.format(base_name, model_name)
        print('registering ' + model_resource_endpoint)
        api.add_resource(
            get_model_resource(model_name, model_),
            model_resource_endpoint,
            model_resource_endpoint + '/',
            endpoint=model_name)

    for var_symbol in model_:
        if type(var_symbol) != UndefinedFunction:
            print('about to register {}'.format(var_symbol))
            register_func_endpoints(api, model_name, model_, base_name, var_symbol)

    # /api/mymodel/evaluate
    evaluate_resource_endpoint = '{}/{}/evaluate'.format(base_name, model_name)
    print('registering ' + evaluate_resource_endpoint)
    api.add_resource(
        get_evaluate_resource(model_name, model_),
        evaluate_resource_endpoint,
        endpoint='/'.join([model_name, 'evaluate'])
        )

    # /api/mymodel/doc
    model_doc_resource_endpoint = '{}/{}/doc'.format(base_name, model_name)
    print('registering ' + model_doc_resource_endpoint)
    api.add_resource(
        get_model_doc_resource(model_name, model_),
        model_doc_resource_endpoint,
        endpoint='/'.join([model_name, 'doc'])
        )

def get_model_doc_resource(model_name, model):
    '''retrieve documentation string for the model'''
    class model_doc_resource(Resource):
        def get(self):
            """get method for the doc string"""
            return model.__doc__
    return model_doc_resource

def get_model_resource(model_name, model):
    """Retrieve resource asscociated with this model"""

    class model_resource(Resource):
        """Resource for this model"""
        def get(self):
            """Get method for this model"""
            model_detail_dict = model.detail().astype(str).to_dict(orient='index')
            return model_detail_dict

    return model_resource

def get_func_resource(model_name, model, var_symbol):
    """Get resource associated with this function"""
    parser = reqparse.RequestParser()
    func = model[var_symbol]
    defaults = get_defaults(func)

    for arg in getfullargspec(func).args:
        if arg in defaults:
            parser.add_argument(arg, type=str)
        else:
            parser.add_argument(arg, type=str, required=True)

    class FuncResource(Resource):
        """Resource associated with this function"""

        def get(self):
            """get method for this resource"""
            args_ = parser.parse_args(strict=True)
            args = dict()

            for argname, val_ in args_.items():
                args[argname] = json.loads(val_, object_hook=deserialize)
                if isinstance(args[argname], str):
                    args[argname] = json.loads(val_, object_hook=deserialize)
                if isinstance(args[argname], list):
                    args[argname] = np.array(args[argname])
            print('{} {} passed {}'.format(
                model_name,
                var_symbol,
                ' '.join(['{}:{}'.format(arg, type(arg_value)) for arg, arg_value in args.items()])))
            result = func(**args)
            print('{} {} function returned {}'.format(
                model_name, var_symbol, type(result)))
            return serialize(result)

            # try:
            #     return result.tolist()
            # except:
            #     return result
    return FuncResource


def get_defaults_resource(model_name, model, var_symbol):
    """Get resource associated with this function's defaults"""

    # These defaults may change if the user redefines a variables
    # Therefore, the defaults need to be checked every time there's a get request

    class DefaultsResource(Resource):
        """Resource associated with this function's defaults"""
        def get(self):
            """get method for this resource"""
            app.logger.info('getting defaults for {}.{}'.format(model_name, var_symbol))
            func = model[var_symbol]
            function_defaults = get_defaults(func)
            try:
                function_defaults_ = json.dumps(function_defaults, default=serialize)
            except:
                print('problem with {}.{}'.format(model_name, var_symbol))
                raise
            return function_defaults_

    return DefaultsResource

def get_data_resource(model_name, model, var_symbol):
    """Get resource associated with this function's data"""

    # The data may change if the user redefines a variable
    # Data needs to come from the current state of the model

    class DefaultsResource(Resource):
        """Resource associated with this function's defaults"""
        def get(self):
            """get method for this resource"""
            print('getting data for {}.{}'.format(model_name, var_symbol))
            func = model[var_symbol]

            # assume function is kamodofied
            try:
                func_data = func.data
            except AttributeError:
                raise AttributeError('function {}.{} has no data!'.format(
                    model_name, var_symbol))

            try:
                func_data_ = json.dumps(func_data, default=serialize)
            except:
                print('problem with {}.{} data'.format(model_name, var_symbol))
                raise
            return func_data_

    return DefaultsResource


def get_evaluate_resource(model_name, model):
    """get resource associated with evaluate"""
    # not sure what to do if the model changes
    # get doesn't know what variables exist in the model
    # maybe move to route/<model_name>

    parser = reqparse.RequestParser()
    parser.add_argument('variable', type=str, required=True)

    for var_symbol in model:
        if type(var_symbol) != UndefinedFunction:
            func = model[var_symbol]
            for arg in getfullargspec(func).args:
                parser.add_argument(arg, type=str)


    class EvaluateResource(Resource):
        """Resource associated with evaluate"""

        def get(self):
            """get method for this resource"""
            args_ = parser.parse_args()
            args = dict()

            variable_name = ''

            for argname, val_ in args_.items():
                if argname != 'variable':
                    if val_ is not None:
                        args[argname] = pd.read_json(StringIO(val_), typ='series')

                else:
                    variable_name = val_

            try:
                result = model.evaluate(variable=variable_name, **args)
            except SyntaxError as m:
                return {'message': '{}'.format(m)}

            return {k_: v_.tolist() for k_, v_ in result.items()}

    return EvaluateResource


def get_func_plot_resource(model, var_symbol):
    """Get resource associated with this function's plot"""

    parser = reqparse.RequestParser()
    func = model[var_symbol]
    defaults = get_defaults(func)

    for arg in getfullargspec(func).args:
        if arg in defaults:
            parser.add_argument(arg, type=str)
        else:
            parser.add_argument(arg, type=str, required=True)


    class FuncPlotResource(Resource):
        """Resource associated with evaluate"""
        def get(self):
            """get method for this resource"""
            app.logger.info('function plot resource called')
            args_ = parser.parse_args(strict=True)
            args = dict(indexing='ij') # needed for gridded data

            for argname, val_ in args_.items():
                if val_ is not None:
                    args[argname] = pd.read_json(StringIO(val_), typ='series')
            figure = model.figure(variable=str(type(var_symbol)), **args)
            return go.Figure(figure).to_json()

    return FuncPlotResource

def get_func_doc_resource(model, var_symbol):
    """Get resource associated with this function's documentation"""
    class FuncDocResource(Resource):
        """Resource associated with doc"""
        def get(self):
            """get method for this resource"""
            return model[var_symbol].__doc__
    return FuncDocResource

if __name__ == '__main__':
    main()
